
def make_autograd_function(module):
        RkMod = module

        # -> so we can access to it inside the following class definition
        #  (when defining a Class inside a Class we cannot use `self`)
        class RK_autograd_Function(torch.autograd.Function):
            # === OUR FORWARD FUNCTION ===
            @staticmethod
            def forward(ctx, dummy_input, *args):
                if RkMod.compiler.storage.ld != {}:
                    ctx.RK_Storage = storage = RkMod.compiler.storage
                    ctx.name_of_inputs_which_req_grad = (
                        RkMod.name_of_inputs_which_req_grad_buffer
                    )
                    with torch.enable_grad():
                        exec(RkMod.init_code, RkMod.gd, storage.ld)  # is compiler.gd
                        # for l in RkMod.fwd_fct_list:
                        #     RkMod._exec(l)
                        for op in self.op_list[:self.op_sched.loss_idx]:
                            self._exec(op)
                else:
                    # *** INITIALIZATION PART ***
                    #  -> Get the inputs using the buffer (Rem 1)
                    dict_inputs = RkMod.dict_inputs_buffer
                    RkMod.dict_inputs_buffer = None
                    #  -> Create the RK_Storage for this run, and store it in ctx
                    ctx.RK_Storage = storage = RK_Storage()
                    storage.init(RkMod.gd)
                    RkMod.compiler.storage = storage
                    #  -> Store what we need to return inputs' grad (Rem 1)
                    ctx.name_of_inputs_which_req_grad = (
                        RkMod.name_of_inputs_which_req_grad_buffer
                    )
                    # RkMod.name_of_inputs_which_req_grad_buffer = None
                    #  -> Detach input tensors (Rem 3) and store all the inputs
                    dict_input_tensors_detach = dict()  #  dict : input -> detached input
                    for k, v in dict_inputs.items():
                        if isinstance(v, torch.Tensor):
                            v_d = v.detach().requires_grad_(v.requires_grad)
                            dict_input_tensors_detach[v] = v_d
                            storage.ld[k] = v_d
                        #  TODO elif iterables of Tensors ?
                        else:
                            storage.ld[k] = v
                    

                    torch.cuda.synchronize()
                    with torch.enable_grad():
                        self.init_fwd_exec()
                    torch.cuda.synchronize()

                #  *** EXECUTION PART ***
                # -> Autograd turns off itself before giving use the control.
                # -> But we need it to forward/backward each node.
                
                # -> Get the output
                # outs = [
                #     RkMod.compiler.get_val(out_node.main_target).detach().requires_grad_()
                #     for out_node in RkMod.rkgb_res.forward_and_backward_graph.list_output_data_anodes
                # ]

                outs = []
                for out_node in self.rkgb_res.forward_graph.output_nodes[:1]:
                    # print(anode)
                    RkMod.compiler.get_val(f"out_{out_node.main_target}").data = RkMod.compiler.get_val(out_node.main_target)
                    o = RkMod.compiler.get_val(f"out_{out_node.main_target}")#.detach().requires_grad_()
                    # print(o.grad_fn)
                    outs.append(o)

                if len(outs) == 1:
                    return outs[0]
                else:
                    return tuple(outs)
                # -> Remember that out have been detached from the rest during exec
                """
                ctx.set_materialize_grads(True) # as the default
                # -> so we don't have to check if grad_output is None
                # -> if outputs' grad is None Autograd fill them with zeros
                """

            # === END OF FORWARD FUNCTION ===

            """
            @staticmethod
            def setup_context(ctx,inputs,outputs):
                pass
            # PyTorch prefer to handle the ctx in a separate method,
            # but it makes more sense for us to handle ctx during the forward.
            # (Also, setup_context can only access to inputs and outputs,
            # so they suggest to pass intermediate values as outputs,
            # but outputs can only be tensors)
            # Anyway, it's not important, Autograd is originally designed
            # to handle the ctx during the forward, it's just that PyTorch 2.0
            # now prefers to use this separate method.
            """

            # === OUR BACKWARD FUNCTION ===
            @staticmethod
            @torch.autograd.function.once_differentiable
            def backward(ctx, *grad_outs):  #  TODO multiple outputs
                #  -> Reload the storage and out
                storage = ctx.RK_Storage
                RkMod.compiler.storage = storage
                # -> Put grad_out in out.grad (Rem 4)
                # print(grad_outs)
                # for out_mt, out_grad in zip(RkMod.rkgb_res.S_graph.outputs, grad_outs):
                for out_node, out_grad in zip(
                    self.rkgb_res.forward_and_backward_graph.list_output_data_anodes,
                    grad_outs):
                    out_mt = out_node.main_target
                    out = RkMod.compiler.get_val(out_mt)
                    # out.grad = out_grad.view(out.shape)
                    out.grad = out_grad.data.as_strided_(
                                out.shape, out.stride(), out.storage_offset()
                            )
                    out_grad.data = torch.empty(0)
                # for out_node in self.rkgb_res.forward_graph.output_nodes:
                #         # print(anode)
                #         RkMod.compiler.get_val(f"out_{out_node.main_target}").data = torch.empty(0)
                    
                #  * record_mem stuff *
                if RkMod.exec_with_record_mem:
                    RkMod.output_size = tensor_memory_size(
                        storage.ld[RkMod.output.main_target]
                    )
                    loss_idx = len(RkMod.allo_mem)
                    # self.allo_mem[-1] += self.output.info.memsize
                    # output grad is generated outside
                # -> exec bwd_fct_list with early stop or not
                stop = RkMod.backward_stop
                if stop:
                    len_fwd = len(RkMod.fwd_fct_list)
                    for l in RkMod.bwd_fct_list[: (stop - len_fwd)]:
                        with torch.enable_grad():
                            RkMod._exec(l)
                else:
                    # for l in RkMod.bwd_fct_list:
                    #     with torch.enable_grad():
                    #         RkMod._exec(l)
                    for op in RkMod.op_list[RkMod.op_sched.loss_idx+1:]:
                        RkMod._exec(op)
                    if RkMod.exec_with_record_mem and RkMod.backward_add_output_grad:
                        RkMod.allo_mem[loss_idx] += RkMod.output_size
                    #  -> return grad of dummy input + inputs' which req grad (Rem 1)
                    grad_inputs = tuple(
                        RkMod.compiler.get_val(inp).grad
                        for inp in ctx.name_of_inputs_which_req_grad
                    )
                    grads = (torch.ones(1),) + grad_inputs
                    #  -> Clear the compiler (and Autograd clears ctx)
                    # RkMod.compiler.storage = None
                    for out_node in self.rkgb_res.forward_graph.output_nodes:
                        # print(anode)
                        RkMod.compiler.get_val(f"out_{out_node.main_target}").data = torch.empty(0)
                        
                    return grads

            # === END OF BACKWARD FUNCTION ===

        return RK_autograd_Function



import torch
from rkgb.lowlevel.measure import TimerCPU
from copy import deepcopy

def get_optimize_metrics(
    _p, optim, cpu_optim=None, optim_kwargs={}, niter=20, minor_offload_size=1024**2
):
    # x = torch.ones(30, 1024, 1024, 1024)
    if optim is None:
        return None
    if cpu_optim is None:
        cpu_optim = optim
    # timer = irotor.make_timer(torch.device("cpu"))
    timer = TimerCPU()
    # p = deepcopy(_p).to("cuda")
    # if not p.is_leaf:
    p = torch.ones([10, 1024, 1024], dtype=_p.dtype).to("cuda")
    size = p.numel()
    
    p_c = torch.nn.Parameter(torch.zeros_like(p, device="cpu"))
    p_c.grad = torch.ones_like(p_c)
    optimizer = cpu_optim([p_c], **optim_kwargs)
    for i in range(niter):
        optimizer.step()
    
    torch.cuda.synchronize()

    a_c = torch.ones_like(p, device="cpu", pin_memory=True)
    a_g = torch.ones_like(p, device="cuda")
    b_c = torch.ones_like(p, device="cpu", pin_memory=True)
    b_g = torch.ones_like(p, device="cuda")
    p_stream = torch.cuda.Stream()
    o_stream = torch.cuda.Stream()
    timer.start()
    for i in range(niter):
        with torch.cuda.stream(p_stream):
            a_c.copy_(a_g, non_blocking=True)
            a_c.copy_(a_g, non_blocking=True)
        with torch.cuda.stream(o_stream):
            b_g.copy_(b_c, non_blocking=True)
            b_g.copy_(b_c, non_blocking=True)
        optimizer.step()
        p_c.grad.zero_()
    torch.cuda.synchronize()
    timer.end()
    cpu_optimize_speed = size * p.element_size() * niter / timer.elapsed()
    
    
    p.grad = torch.ones_like(p)
    optimizer = optim([p], **optim_kwargs)
    torch.cuda.reset_peak_memory_stats()
    mem = torch.cuda.memory_allocated()
    optimizer.step()
    timer.start()
    for i in range(niter):
        optimizer.step()
    torch.cuda.synchronize()
    timer.end()
    
    mem_after = torch.cuda.memory_allocated()
    gpu_optimize_speed = size * p.element_size() * niter / timer.elapsed()
    # gpu_optimize_speed = 0
    opt_size = mem_after - mem
    opt_overhead = torch.cuda.max_memory_allocated() - mem_after

    
    optimize_metrics = {
        "cpu_optimize_speed": cpu_optimize_speed * 0.8,
        "gpu_optimize_speed": gpu_optimize_speed,
        "optimizer_states_factor": round(opt_size // size / p.element_size()),
        "optimizer_overhead": round(opt_overhead // size / p.element_size()),
        "cpu_optim": cpu_optim,
        "gpu_optim": optim,
        "optim_kwargs": optim_kwargs,
    }
    return optimize_metrics


def measure_bandwidth(niter=10):
    timer = TimerCPU()
    a_c = torch.ones([10, 1024, 1024], device="cpu", pin_memory=True)
    a_g = torch.ones([10, 1024, 1024], device="cuda")
    b_c = torch.ones([10, 1024, 1024], device="cpu", pin_memory=True)
    b_g = torch.ones([10, 1024, 1024], device="cuda")
    p_stream = torch.cuda.Stream()
    o_stream = torch.cuda.Stream()
    timer.start()
    for i in range(niter):
        with torch.cuda.stream(p_stream):
            a_c.copy_(a_g, non_blocking=True)
        with torch.cuda.stream(o_stream):
            b_g.copy_(b_c, non_blocking=True)
    torch.cuda.synchronize()
    timer.end()
    bandwidth = niter * a_c.numel() * a_c.element_size() / timer.elapsed()
    return bandwidth


from rkgb.core.base import Node
from rkgb.core.backward import ComputationNode, AllocationNode
from rkgb.core.hierarchical import HierarchicalCluster
from collections import namedtuple
from copy import deepcopy
from typing import List
import warnings
import numpy as np
import torch
from rkgb.lowlevel.ast_add_on import make_str_list_assign
from .op_schedule import (
    Allocation,
    Activation,
    Parameter,
    Buffer,
    Op,
    ComputeOp,
    DeleteOp,
    MappingOp,
    AllocateOp,
    OffloadOp,
    PrefetchOp,
    SynchronizeOp,
    OptimizeOp,
    OpSchedule,
    ExecCodeOp,
)


class ListOp(list):
    def __init__(self, ops: List[Op], constant_cost=0):
        super(ListOp, self).__init__(ops)
        self._pop = super(ListOp, self).pop
        self._remove = super(ListOp, self).remove
        self._append = super(ListOp, self).append
        self._insert = super(ListOp, self).insert
        self.constant_cost = constant_cost
        self.time = sum(op.time for op in ops)+self.constant_cost

    def pop(self, index):
        self.time -= self[index].time
        return self._pop(index)

    def remove(self, op: Op):
        self.time -= op.time
        return self._remove(op)

    def append(self, op: Op):
        self.time += op.time
        return self._append(op)

    def insert(self, i, op: Op):
        self.time += op.time
        return self._insert(i, op)


class Step:
    def __init__(self, op_list: List[Op]) -> None:
        main_op_list = [op_list[0]]
        self_op_list = []
        self.del_ops = []

        sync = False
        for op in op_list[1:]:
            if isinstance(op, SynchronizeOp):
                sync = True
            if not sync:
                main_op_list.append(op)
            else:
                if isinstance(op, DeleteOp): 
                    self.del_ops.append(op)
                else:
                    self_op_list.append(op)

        ofl_ops = []
        prf_ops = []
        prf_act_ops = []
        ofl_act_ops = []
        self_del_list = []
        opt_ops = []
        comp_ops = []
        self.alloc_ops = []
        self.cpu_constant_cost = 50
        
        for op in main_op_list:
            if (isinstance(op, SynchronizeOp)) or isinstance(
                op, AllocateOp
            ):
                self.alloc_ops.append(op)
            elif isinstance(op, OffloadOp):
                if isinstance(op.target, Parameter):
                    ofl_ops.append(op)
                else:
                    ofl_act_ops.append(op)
            elif isinstance(op, PrefetchOp):
                if isinstance(op.target, Parameter):
                    prf_ops.append(op)
                else:
                    prf_act_ops.append(op)
            elif isinstance(op, OptimizeOp) and op.is_cpu:
                opt_ops.append(op)
            elif isinstance(op, DeleteOp) and isinstance(op.target, Parameter):
                self.del_ops.append(op)
            elif isinstance(op, DeleteOp) and op.wait_events:
                self_del_list.append(op)
            else:
                comp_ops.append(op)

        self.ofl_ops = ListOp(ofl_ops + ofl_act_ops)
        self.prf_ops = ListOp(prf_act_ops + prf_ops)
        self.opt_ops = ListOp(opt_ops, self.cpu_constant_cost)
        self.comp_ops = ListOp(comp_ops)
        self.self_ops = ListOp(self_del_list+self_op_list)
        self.op_time = self.simulate_schedule()

    @property
    def op_list(self):
        if not self.opt_ops:
            opt_ops = []
        else:
            list_params = [p for op in self.opt_ops for p in op.list_params]
            opt_ops = self.opt_ops
        return (
            self.alloc_ops
            + self.prf_ops
            + self.comp_ops
            + self.ofl_ops
            + self.self_ops
            + opt_ops
            + self.del_ops
        )

    @property
    def time(self):
        # return max(
        #     max(
        #     self.ofl_ops.time, 
        #     self.comp_ops.time
        # ) + self.self_ops.time,
        #     self.prf_ops.time, 
        #     self.opt_ops.time)
        return max(self.op_time.values())
    
    def simulate_schedule(self):
        op_time = {}
        time = 0
        for op in self.prf_ops:
            op_time[op.name] = time
            time += op.time

        time = 0
        for op in self.comp_ops:
            wait_time = [op_time[f"{e[0]}({e[1]})"] 
                           for e in op.wait_events 
                           if f"{e[0]}({e[1]})" in op_time]
            if wait_time:
                time = max(wait_time+[time])
            time += op.time
            op_time[op.name] = time
        
        time = 0
        for op in self.ofl_ops:
            wait_time = [op_time[f"{e[0]}({e[1]})"] 
                           for e in op.wait_events 
                           if f"{e[0]}({e[1]})" in op_time]
            if wait_time:
                time = max(wait_time+[time])
            time += op.time
            op_time[op.name] = time
        
        time = self.cpu_constant_cost
        for op in self.opt_ops:
            wait_time = [op_time[f"{e[0]}({e[1]})"] 
                           for e in op.wait_events 
                           if f"{e[0]}({e[1]})" in op_time]
            if wait_time:
                time = max(wait_time+[time])
            time += op.time
            op_time[op.name] = time
        return op_time

    def all_time(self):
        return (
            self.ofl_ops.time,
            self.prf_ops.time,
            self.opt_ops.time,
            self.comp_ops.time,
        )

    def max2nd(self):
        t = list(self.all_time())
        t.remove(max(t))
        return max(t)

    @property
    def max_comp_time(self):
        return self.op_time[self.comp_ops[-1].name]


class AliveSimulator:
    """
    Dynamic alive status of different allocations.
    Functions:
    1. Easy to check if one allocation is alive for a range of steps.
    2. Easy to update when operations are changed.
    3. Easy to translate allocations for anonymized cluster.
    4. Return peak memory fast.
    Assumption:
    1. Number of operations will not be changed (operations can be disabled).
    2. Number of allocations will not be changed.
    """

    def __init__(self, alive_list, dict_alloc, alloc_categories={}):
        self.update_dict_alloc(dict_alloc)
        self.length = len(alive_list)
        self.alloc_mem_np = np.array(
            [self.dict_alloc[alloc_name].mem for alloc_name in self.alloc_names]
        )
        self.alive_np = np.array(
            [
                [
                    1 if alive_list[i][alloc_name] else 0
                    for alloc_name in self.alloc_names
                ]
                for i, _ in enumerate(alive_list)
            ],
            dtype=bool,
        )

        self.categories = {}
        self.categories["activation"] = np.array(
            [
                self.alloc_names.index(alloc_name)
                for alloc_name in self.alloc_names
                if isinstance(self.dict_alloc[alloc_name], Activation)
            ]
        )
        self.categories["parameter"] = np.array(
            [
                self.alloc_names.index(alloc_name)
                for alloc_name in self.alloc_names
                if isinstance(self.dict_alloc[alloc_name], Parameter)
            ]
        )
        self.categories["optim_states"] = np.array(
            [
                self.alloc_names.index(alloc_name)
                for alloc_name in self.alloc_names
                if isinstance(self.dict_alloc[alloc_name], Parameter)
                and self.dict_alloc[alloc_name].is_optim_states
            ]
        )
        for k, v in alloc_categories.items():
            self.categories[k] = np.array(
                [self.alloc_names.index(alloc_name) for alloc_name in v]
            )

    def update_dict_alloc(self, dict_alloc):
        """
        For translation: need to ensure the dict_alloc to appear in the same list
        """
        self.dict_alloc = dict_alloc
        self.alloc_names = [alloc_name for alloc_name in sorted(dict_alloc.keys())]
        self.name2key = {alloc_name: i for i, alloc_name in enumerate(self.alloc_names)}

    def _alive(self, key: int, idx_range: range):
        if isinstance(idx_range, int):
            idx_range = range(idx_range, idx_range + 1)
        return self.alive_np[idx_range, key]
  
    def _mem(self, idx_range: range, category=None):
        """
        Return a (idx_range,)-shape array of memory over the steps range.
        """
        if category:
            if isinstance(idx_range, int):
                idx_range = range(idx_range, idx_range + 1)
            return np.sum(
                (self.alive_np[idx_range] * self.alloc_mem_np)[
                    :, self.categories[category]
                ],
                axis=1,
            )
        return np.matmul(self.alive_np[idx_range], self.alloc_mem_np)

    def act_mem(self, idx_range, category="activation"):
        return self._mem(idx_range, category=category)

    def param_mem(self, idx_range, category="parameter"):
        return self._mem(idx_range, category=category)

    def save_mem(self, idx_range=None, act_multiplier=1):
        if idx_range is None:
            idx_range = range(0, self.length)
        return act_multiplier * self.act_mem(idx_range) + self.param_mem(idx_range)

    def alive_once(self, key: int, idx_range: range):
        # being alive AT LEAST once
        if isinstance(idx_range, int):
            idx_range = range(idx_range, idx_range + 1)
        if isinstance(key, str):
            key = self.name2key[key]
        return bool(max(self._alive(key, idx_range)))

    def alive_always(self, key: int, idx_range: range):
        # being alive AT LEAST once
        if isinstance(idx_range, int):
            idx_range = range(idx_range, idx_range + 1)
        if isinstance(key, str):
            key = self.name2key[key]
        return bool(min(self._alive(key, idx_range)))


class Simulator:
    """
    With steps and np-alive status, simulator allows efficient
    simulation of swapping/changing operations. The final output
    of operation list can be used in op_sched updatation.
    """

    def __init__(self, op_sched: OpSchedule):
        self.op_sched = op_sched
        self.op_list: List[Op] = op_sched.op_list
        self.loss_idx = op_sched.loss_idx
        alive_list = (
            op_sched.alive_list
            if hasattr(op_sched, "alive_list")
            else op_sched.create_alive_list()
        )
        self.alive_list = AliveSimulator(alive_list, dict_alloc=op_sched.dict_alloc)
        self.create_steps()

    def refine_optimize(self):
        steps = self.steps
        for j in range(len(steps) - 1, self.loss_step, -1):
            opt_ops: List[OptimizeOp] = list(steps[j].opt_ops)
            opt2user_step = {op.name: None for op in opt_ops}
            steps_avail = {op: steps[j:] for op in opt_ops}

            for i, step in enumerate(steps[:]):
                if None not in opt2user_step.values():
                    break
                for opt_op in opt_ops:
                    if opt2user_step[opt_op.name] is not None:
                        continue
                    # for usr in opt_op.target.pnode.users_real:
                    #     if usr.name in [str(op) for op in step.comp_ops]:
                    for prf_op in step.prf_ops:
                        if isinstance(prf_op.target, Activation):
                            continue
                        if prf_op.target.param_name == opt_op.target.pnode.param_name:
                            # print(opt_op.name)
                            opt2user_step[opt_op.name] = i
                            steps_avail[opt_op].extend(steps[:i])
                            # located_ops.append(opt_op)
                            # opt_ops.remove(opt_op)
            for op, avail in steps_avail.items():
                avail_step = max(avail, key=lambda x: x.max_comp_time - x.opt_ops.time)
                # print(avail_step.time,avail_step.opt_ops.time)
                # print(steps[j].time,steps[j].opt_ops.time)
                if avail_step.time < avail_step.opt_ops.time:
                    continue
                if steps[j].opt_ops.time < steps[j].max_comp_time:
                    continue
                if (
                    avail_step.opt_ops.time + op.time - avail_step.max_comp_time
                    > steps[j].opt_ops.time - steps[j].max2nd()
                ):
                    continue

                # print(avail_step.time, avail_step.opt_ops.time)
                avail_step.opt_ops.append(op)
                steps[j].opt_ops.remove(op)
        
        for step in self.steps:
            step.op_time = step.simulate_schedule()
            
        self.refine()
        self.recreate_op_list()

    def create_steps(self):
        self.steps: List[Step] = []
        step_op = []
        for i, op in enumerate(self.op_list):
            if isinstance(op, SynchronizeOp) and op.stream is None:
                if step_op:
                    self.steps.append(Step(step_op))
                step_op = []
            if i == self.loss_idx:
                self.loss_step = len(self.steps)
            step_op.append(op)
        self.steps.append(Step(step_op))
        self.recreate_op_list()

    def recreate_op_list(self):
        self.loss_op = self.op_list[self.loss_idx]
        op_list = []
        for step in self.steps:
            op_list += step.op_list
        self.op_list = op_list
        self.loss_idx = self.op_list.index(self.loss_op)
        self.get_occurrences()

    def get_occurrences(self):
        self.occurrences = dict()
        for i, op in enumerate(self.op_list):
            if op.name in self.occurrences:
                self.occurrences[op.name].append(i)
            else:
                self.occurrences[op.name] = [i]

    def refine(self):
        """
        Disable deletion to avoid infeasible operations in the list.
        """
        for idx, op in enumerate(self.op_list):
            if op.disabled:
                continue
            if "loss" in op.name:
                op.disabled = True
            if isinstance(op, DeleteOp) and isinstance(op.target, Activation):
                used_idx = [
                    i
                    for cnode in op.target.anode.users_real
                    for i in self.occurrences[ComputeOp(cnode).name]
                    if not "loss" in cnode.name
                ]
                if not used_idx or max(used_idx) < idx:
                    continue
                next_used_i = min(i for i in used_idx if i > idx)

                for cnode in op.target.anode.deps:
                    c_op = ComputeOp(cnode)
                    p_op = PrefetchOp(op.target)
                    if not self.op_sched.is_occurred(
                        c_op.name, idx, next_used_i
                    ) and not self.op_sched.is_occurred(p_op.name, idx, next_used_i):
                        op.disabled = True
                        break



from .rockmate import Rockmate
from . import solvers
from rkgb import partitioned
from configmypy import Bunch
import yaml
from copy import copy

import inspect

available_solvers = {
    "hilp": solvers.HILP,
    "rotor": solvers.RK_rotor,
    "cheap": solvers.CheapSolver,
    "twremat": solvers.TwRemat
}

available_partitioners = {
    "sequence": partitioned.PartitionerSequence,
    "bottom_to_top": partitioned.PartitionerBottomToTop,
    "repetitive": partitioned.PartitionerRecognizeRepetitivePattern,
    "base": partitioned.Partitioner,
    "dagp": partitioned.PartitionerDagp,
}

def _default_config(func):
    config = Bunch()
    sign = inspect.signature(func)
    sample = func()
    for arg in sign.parameters.values():
        if arg.default is not arg.empty:
            config[arg.name] = getattr(sample, arg.name)
    return config

def default_config(name):
    if name in available_solvers:
        func = available_solvers[name].Config
    elif name in available_partitioners:
        func = available_partitioners[name].Config
    else:
        raise ValueError(f"default_config: unknown solver or partitioner {name}")
    return _default_config(func)

def _add_solver(config, name, position):
    defaults = _default_config(available_solvers[name].Config)
    config.solver[position][name] = defaults

def _add_top_solver(config, name):
    _add_solver(config, name, "top")
def _add_bottom_solver(config, name):
    _add_solver(config, name, "bottom")

def _add_partitioner(config, name):
    defaults = _default_config(available_partitioners[name].Config)
    config.partitioner[name] = defaults

def generate_config(config_type):
    result = Bunch()
    result.solver = Bunch()
    result.solver.top = Bunch()
    result.solver.bottom = Bunch()
    result.partitioner = Bunch()
    if config_type == "rotor":
        _add_top_solver(result, "rotor")
        _add_partitioner(result, "sequence")
    elif config_type == "rockmate":
        _add_bottom_solver(result, "hilp")
        _add_top_solver(result, "rotor")
        result.solver.bottom.hilp.nb_bdg_save = 10
        result.solver.bottom.hilp.nb_bdg_peak = 10
        _add_partitioner(result, "sequence")
    elif config_type == "checkmate":
        _add_top_solver(result, "hilp")
        result.solver.top.hilp.time_limit = 3600
        result.solver.top.hilp.nb_total_nodes = 10000
    elif config_type == "hilp":
        _add_bottom_solver(result, "hilp")
        _add_top_solver(result, "hilp")
        _add_partitioner(result, "bottom_to_top")
        result.solver.top.hilp.nb_total_nodes = result.partitioner.bottom_to_top.max_estimate_for_main_graph
        result.solver.top.hilp.time_limit *= 20
        result.solver.bottom.hilp.accurate_mem = False
        result.partitioner.bottom_to_top.can_use_rotor = False
    elif config_type == "hiremate":
        for solver in available_solvers.keys():
            _add_top_solver(result, solver)
            _add_bottom_solver(result, solver)
        result.solver.top.hilp.nb_total_nodes = 100
        result.solver.top.hilp.time_limit *= 20
        result.solver.bottom.hilp.accurate_mem = False
        for partitioner in available_partitioners.keys():
            _add_partitioner(result, partitioner)
    elif config_type == "offmate":
        _add_top_solver(result, "hilp")
        _add_bottom_solver(result, "cheap")
        result.solver.top.hilp.offload = True
        result.solver.top.hilp.nb_total_nodes = 100
        result.solver.top.hilp.nb_total_sched = 500
        result.solver.top.hilp.time_limit = 1200
        result.solver.bottom.cheap.add_offload = True
        _add_partitioner(result, "repetitive")
    elif config_type == "noremat":
        _add_top_solver(result, "cheap")
        result.solver.top.cheap.cheap_factor = 1e9
        _add_partitioner(result, "base")
    else:
        raise ValueError(f"Unknown config type {config_type}. Valid values are:"
                         "rotor, rockmate, checkmate, hilp, hiremate")
    return result

_yaml_inited = False
def save_config(config, filename):
    global _yaml_inited
    if not _yaml_inited:
        from yaml.representer import Representer
        yaml.add_representer(Bunch, Representer.represent_dict)
        _yaml_inited = True
    with open(filename, "w") as f:
        yaml.dump(config, f)

def load_config(filename):
    with open(filename, "r") as f:
        result = yaml.load(f, yaml.SafeLoader)
    return Bunch(result)

def _make_solver(name: str, b: Bunch):
    return available_solvers[name](**b)

def _make_partitioner(name: str, b: Bunch):
    return available_partitioners[name](**b)

def from_config(model, model_inputs, budget=None, config_type="hiremate", config=None, **kwargs):
    if config is None:
        config = generate_config(config_type)
    top_solvers = [ _make_solver(name, conf) for name, conf in config.solver.top.items() if conf is not None ]
    bottom_solvers = [ _make_solver(name, conf) for name, conf in config.solver.bottom.items() if conf is not None ]
    partitioners = [ _make_partitioner(name, conf) for name, conf in config.partitioner.items() if conf is not None ]
    if not partitioners:
        partitioners = [ partitioned.Partitioner() ]
    return Rockmate(model, model_inputs, budget, top_solvers=top_solvers, bottom_solvers=bottom_solvers,
                    partitioners=partitioners, **kwargs)

class PureRotor(Rockmate):
    @classmethod
    def solver(cls, force_python=False):
        return solvers.RK_rotor(force_python=force_python)

    @classmethod
    def partitioner(cls):
        return partitioned.PartitionerSequence(sub_partitioner=partitioned.Partitioner()) 

    def __init__(self, model, model_inputs, budget=None, rotor_solve_with_python=False, **kwargs):
        assert "list_solvers" not in kwargs
        assert "partitioners" not in kwargs

        super().__init__(model, model_inputs, budget,
                         top_solvers=[self.solver(force_python=rotor_solve_with_python)],
                         partitioners=[self.partitioner()],
                         **kwargs)

class PureCheckmate(Rockmate):
    @classmethod
    def solver(cls):
        hilp_solver = solvers.HILP(ilp_solver="PULP_CBC_CMD", offload=False)
        return hilp_solver

    def __init__(self, model, model_inputs, budget=None, **kwargs):
        assert "list_solvers" not in kwargs
        assert "partitioners" not in kwargs

        super().__init__(model, model_inputs, budget=budget,
                         top_solvers=[self.solver()],
                         bottom_solvers=[],
                         partitioners=[],
                         **kwargs)

class PureRockmate(Rockmate):

    @classmethod
    def partitioner(cls):
        return PureRotor.partitioner()

    def __init__(self, model, model_inputs, budget=None, rotor_solve_with_python=False, **kwargs):
        assert "list_solvers" not in kwargs
        assert "partitioners" not in kwargs

        hilp_solver = solvers.HILP(ilp_solver="PULP_CBC_CMD", offload=False)
        rk_solver = PureRotor.solver(force_python=rotor_solve_with_python)
        super().__init__(model, model_inputs, budget=budget,
                         top_solvers=[rk_solver],
                         bottom_solvers=[hilp_solver],
                         partitioners=[self.partitioner()],
                         **kwargs)

class Hiremate(Rockmate):
    ## valid_algorithms = ["hilp", "twremat", "rotor"]
    valid_algorithms = ["hilp", "rotor"]

    def __init__(self, model, model_inputs, budget=None, algorithms=valid_algorithms, partitioner_params={}, **kwargs):
        assert "list_solvers" not in kwargs
        assert "partitioners" not in kwargs

        assert all(alg in self.valid_algorithms for alg in algorithms), f"invalid algorithm list '{','.join(algorithms)}'"
        assert "can_use_rotor" not in partitioner_params
        assert "main_graph_as_any_other" not in partitioner_params

        if not ('rotor' in algorithms): # hilp, twremat+hilp
            list_partitioners = [
                partitioned.PartitionerBottomToTop(
                    can_use_rotor=False,
                    **partitioner_params
                )
            ]
        else: # twremat+hilp+rotor
            list_partitioners = [
                partitioned.PartitionerBottomToTop(
                    can_use_rotor=True,
                    **partitioner_params
                ),
                partitioned.PartitionerSequence(
                    partitioned.PartitionerBottomToTop(
                        can_use_rotor=True,
                        main_graph_as_any_other=True,
                        **partitioner_params
                    )
                )
            ]
        
        top_solvers = []
        bottom_solvers = []
        if "hilp" in algorithms:
            top_solvers.append(solvers.HILP(time_limit=600, nb_total_nodes=100))
            bottom_solvers.append(solvers.HILP(time_limit=60))
        if "rotor" in algorithms:
            top_solvers.append(solvers.RK_rotor())
            bottom_solvers.append(solvers.RK_rotor())
        if "twremat" in algorithms:
            top_solvers.append(solvers.TwRemat())
            bottom_solvers.append(solvers.TwRemat())

        super().__init__(model, model_inputs, budget=budget,
                         top_solvers=top_solvers,
                         bottom_solvers=bottom_solvers,
                         partitioners=list_partitioners,
                         **kwargs)
        
# TODO: Careful, Offmate does not have the same assumptions about 1/
# optimizer 2/ where the model is at the start 3/ do we handle
# parameter gradients or leave them untouched in memory
class Offmate(Rockmate):
    valid_algorithms = ["hilp"]

    def __init__(self, model, model_inputs, nlayers=32, budget=None, algorithms=valid_algorithms, partitioner_params={}, **kwargs):
        assert "list_solvers" not in kwargs
        assert "partitioners" not in kwargs

        assert all(alg in self.valid_algorithms for alg in algorithms), f"invalid algorithm list '{','.join(algorithms)}'"
        # assert "can_use_rotor" not in partitioner_params
        assert "main_graph_as_any_other" not in partitioner_params

        # if not ('rotor' in algorithms): # hilp, twremat+hilp
        #     list_partitioners = [
        #         partitioned.PartitionerBottomToTop(
        #             can_use_rotor=False,
        #             **partitioner_params
        #         )
        #     ]

        partitioners = [partitioned.PartitionerRecognizeRepetitivePattern(
            strict_max_number_of_top_level_nodes=nlayers+4,
            max_number_of_patterns=nlayers+2,
            min_percentage_covered_required=0.75)]
        # else: # twremat+hilp+rotor
        #     list_partitioners = [
        #         partitioned.PartitionerBottomToTop(
        #             can_use_rotor=True,
        #             **partitioner_params
        #         ),
        #         partitioned.PartitionerSequence(
        #             partitioned.PartitionerBottomToTop(
        #                 can_use_rotor=True,
        #                 main_graph_as_any_other=True,
        #                 **partitioner_params
        #             )
        #         )
        #     ]
        
        solver = solvers.HILP(ilp_solver="PULP_CBC_CMD")
        solver.config.offload = True
        solver.config.solve_only_top_level = True
        top_solvers = [solver]
        bottom_solvers = [solvers.CheapSolver()]
        # if "hilp" in algorithms:
        #     top_solvers.append(solvers.HILP(time_limit=600, nb_total_nodes=100))
        #     bottom_solvers.append(solvers.HILP(time_limit=60))
        # if "rotor" in algorithms:
        #     top_solvers.append(solvers.RK_rotor())
        #     bottom_solvers.append(solvers.RK_rotor())

        super().__init__(model, model_inputs, budget=budget,
                         top_solvers=top_solvers,
                         bottom_solvers=bottom_solvers,
                         partitioners=partitioners,
                        #  gpu_optim=torch.optim.Adam,
                         minor_offload_size=10*1024**2,
                         **kwargs)



import time
import torch
import rkgb
import numpy as np
from rkgb.core.hierarchical import HierarchicalGraph, HierarchicalCluster
import gc
from ..main import (
    Solver,
    get_cluster_budget,
    get_hgraph_budget_lb,
    get_hgraph_budget_ub,
)
import pulp
from rkgb.core.hierarchical import HierarchicalGraph, HierarchicalCluster
from rkgb.lowlevel.constants import init_target_string
from ...op_schedule import *
from .ilp_model import ModelPULP
from .ilp_offload import ModelPULPOffload
from .ilp_schedule import schedule
from .ilp_utils import (
    set_hcn_list_sched,
    set_hg_parameter_groups,
    clean_hcn_list_sched,
    clean_hg_parameter_groups,
)
import psutil
from dataclasses import dataclass, field


class HILP(Solver):
    '''ILP-based rockmate solver. Generic implementation, can be used in rematerialization or offload mode

    The original idea for the ILP formulation is from Checkmate: https://github.com/parasj/checkmate
    described in

    Paras Jain, Ajay Jain, Aniruddha Nrusimha, Amir Gholami, Pieter Abbeel, Kurt Keutzer, Ion Stoica,
    Joseph E. Gonzalez. Checkmate: Breaking the Memory Wall with Optimal Tensor Rematerialization. MLSys
    2020, https://arxiv.org/abs/1910.02653

    This version is heavily modified from this original implementation, adding a hierarchical approach
    and many features to support activation and parameter offloading.
    '''
    @dataclass
    class Config:
        '''Configuration options for HILP solver'''
        mem_unit: int = 1024**2
        '''How memory values (in bytes) are quantized to obtain reasonable integers'''
        ilp_solver_params: dict = field(default_factory=lambda:
                                        {
                                            "LogToConsole": 0,
                                            "IntegralityFocus": 1,
                                            "NodeFileStart": 0.5,
                                        })
        '''Arguments passed to the ILP solver directly'''
        ilp_solver: str = "PULP_CBC_CMD"
        '''Which solver to use in the PuLP library.'''
        protected_names: list = field(default_factory=list)
        nb_total_sched: int = 100
        '''Total number of options to consider when solving one level of ILP.'''
        nb_total_nodes: int = 20
        nb_bdg_save: int = 6
        nb_bdg_peak: int = 4
        time_limit: int = 1*60
        offload: bool = False ## Only to be used in top solver
        add_offload_sched: bool = False ## Only use in top solver, to automatically add offload versions of computed schedules
        activation_offload: bool = False
        model_kwargs: dict = field(default_factory=dict) ## Passed to the ModelILP class
        accurate_mem: bool = True ## If True, include correction terms
        minor_offload_size:int = 10*1024**2

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.ilp_solver = self.config.ilp_solver
        try:
            solver = pulp.get_solver(self.ilp_solver, msg=0)
        except:
            avail_solver = pulp.listSolvers(onlyAvailable=True)[0]
            print(f"Cannot get {self.ilp_solver}, will use {avail_solver}")
            self.ilp_solver = avail_solver
        print(f"Using {self.ilp_solver} to solve ILP")

    # def __repr__(self):
    #     return f"HILP solver"

    def can_solve(self, hg: HierarchicalGraph):
        return len(hg.list_HCNs) // 2 <= self.config.nb_total_nodes

    def get_budget_list(self, hgraph: HierarchicalGraph):
        min_bdg = get_hgraph_budget_lb(hgraph)
        max_bdg = get_hgraph_budget_ub(hgraph)
        # interfaces_mem = sum(kdn.mem for kdn in hgraph.cluster.all_interfaces)
        interfaces_mem = sum(
            kdn.mem for kdn in hgraph.cluster.interfaces["input_data_anodes"]
        )
        interfaces_mem += sum(
            kdn.mem for kdn in hgraph.cluster.interfaces["output_data_anodes"]
        )

        budgets = []
        l_bd_peak = (
            np.linspace(min_bdg, max_bdg, self.config.nb_bdg_peak) + interfaces_mem
        )
        for bd_peak in l_bd_peak:
            l_bd_save = np.linspace(
                interfaces_mem,
                # min(bd_peak, autograd_sched.mem),
                bd_peak,
                self.config.nb_bdg_save,
            )

            budgets.append((bd_peak, l_bd_save))
        return budgets

    def _group_parameters(self, hg: HierarchicalGraph, minor_size=10*1024**2):
        all_params = {}
        param2hcn = {}

        for i, hcn in enumerate(hg.list_HCNs):
            if not hasattr(hcn, "required_parameter_nodes_real"):
                continue
            if hcn.sub_cluster is not None and hasattr(
                hcn.sub_cluster, "parameter_nodes"
            ):
                # if FWD/BWD hcns have different req_pnodes, parameters may be needed for recomputation
                req_pnodes = hcn.sub_cluster.parameter_nodes
            else:
                h_pnodes = (
                    hcn.required_parameter_nodes_real
                    | hcn.required_parameter_nodes_fake
                )
                req_pnodes = [pnode.original_param_node for pnode in h_pnodes]
            for pnode in req_pnodes:
                if pnode.is_buffer:
                    continue
                if pnode.mem<minor_size:
                    continue
                if pnode not in param2hcn:
                    param2hcn[pnode] = {i}
                else:
                    param2hcn[pnode].add(i)

                # # sub_c2params[sub_cluster.name].add(pnode.param_name)
                # if hasattr(pnode, "original_param_node"):
                #     all_params[pnode.param_name] = pnode.original_param_node
                # else:
                #     all_params[pnode.param_name] = pnode
                # if pnode.param_name not in param2hcn:
                #     param2hcn[pnode.param_name] = {i}
                # else:
                #     param2hcn[pnode.param_name].add(i)

        parameter_groups = {}
        for p, c in param2hcn.items():
            c_ = tuple(sorted(c))
            if c_ not in parameter_groups:
                parameter_groups[c_] = {p}
            else:
                parameter_groups[c_].add(p)

        set_hg_parameter_groups(hg, parameter_groups)

        # parameters = []
        # param_group2hcn = {}
        # for hcn, v in result.items():
        #     param_group2hcn[len(parameters)] =
        #     parameters.append([all_params[p] for p in v])
        # return param_group2hcn, parameters

    def _select_sched(self, hg, overall_budget=None):
        # for fwd hcn, select sched from hcn.sub_cluster and put in hcn.list_sched
        weights = []
        overall_budget = overall_budget or np.inf
        for hcn in hg.list_HCNs:
            if hcn.is_fwd:
                if hcn.sub_cluster is None:
                    weights.append(0)
                else:
                    weights.append(len(hcn.sub_cluster.list_cnodes))

        for i, (hcn, w) in enumerate(zip(hg.list_HCNs, weights)):
            # if i<len(hg.list_HCNs)//2-2 and hcn.sub_cluster is not None:
            #     hcn.list_sched = [min(hcn.sub_cluster.representee_cluster.list_schedules,
            #                key=lambda x: x.mem)]
            #     continue
            nb_sched = max(
                self.config.nb_total_sched * w // sum(weights), 1
            )  # at least 1 sched
            # nb_sched = 3 if i < 30 else nb_sched
            if hcn.sub_cluster is not None:
                # list_sched = hcn.sub_cluster.get_sched(pareto=True)
                list_sched = hcn.sub_cluster.representee_cluster.list_schedules
                list_sched = [
                    op_sched
                    for op_sched in list_sched
                    if op_sched.mem <= overall_budget
                ]
                if nb_sched >= len(list_sched):
                    # hcn.list_sched = list_sched
                    set_hcn_list_sched(hcn, list_sched)
                    if self.config.add_offload_sched:
                        self.add_offload_sched_hcn(hcn)
                    continue
                indices = np.array(
                    [(i, op_sched.mem) for i, op_sched in enumerate(list_sched)]
                )
                sel_sched = [list_sched[0]]
                sel_mem = [list_sched[0].mem]

                while len(sel_sched) < nb_sched:
                    # add the one with most different .mem with all selected sched
                    if (
                        np.max(
                            [min(abs(x - y) for y in sel_mem) for x in indices[:, 1]]
                        )
                        == 0
                    ):  # no different schedules:
                        break
                    argmax_diff = np.argmax(
                        [min(abs(x - y) for y in sel_mem) for x in indices[:, 1]]
                    )
                    sel_mem.append(indices[argmax_diff][1])
                    sel_sched.append(list_sched[argmax_diff])
                    indices[argmax_diff][1] = 0
                hcn.list_sched = sel_sched
                if self.config.add_offload_sched:
                    self.add_offload_sched_hcn(hcn)
                # hcn.list_sched = list_sched[:nb_sched]
            else:
                hcn.list_sched = []
            


    def solve(
        self,
        cluster: HierarchicalCluster,
        budgets=None,
        gc_collect=True,
    ):
        print(f"solving {cluster.name} with budgets {budgets}")
        list_op_sched = []

        if self.config.offload:
            self.model_ilp = ModelPULPOffload
        else:
            self.model_ilp = ModelPULP

        # for hg in cluster.representee_cluster.possible_hg:
        for hg in cluster.representee_cluster.partitionings:
            if budgets is None:
                # self.budgets = get_cluster_budget(
                #     cluster.representee_cluster, with_save_budget=True
                # )
                self.budgets = self.get_budget_list(hg)
            else:
                self.budgets = budgets
            # mem = psutil.virtual_memory()
            # print(
            #         f"The CPU mem usage before solving {cluster.name} is: ",
            #         # psutil.cpu_percent(4)
            #         mem.used,
            #     )
            for budget in self.budgets:
                if not hasattr(budget, "__iter__"):
                    budget = [budget]
                list_op_sched.extend(
                    self.solve_hg(
                        hg,
                        *budget,
                    )
                )
        if gc_collect:
            gc.collect()
        return list_op_sched

    def solve_hg(
        self,
        hg: HierarchicalGraph,
        peak_budget,
        save_budget=None,
        print_result=False,
    ):
        # print(accurate_mem)
        gc.collect()
        if not self.can_solve(hg):
            return []
        if save_budget is not None:
            save_budget = save_budget
        else:
            save_budget = peak_budget

        list_op_sched = []
        self._select_sched(hg, overall_budget=peak_budget)
        
        if not hasattr(save_budget, "__iter__"):
            save_budget = [save_budget]
        # start = time.time()
        if self.config.offload:
            self._group_parameters(hg, minor_size=self.config.minor_offload_size)
        ilp_solver_params = dict(self.config.ilp_solver_params)
        ilp_solver_params["TimeLimit"] = self.config.time_limit

        def solve_md(
            ilp_solver_params=ilp_solver_params,
            model_ilp=self.model_ilp,
            protected_names=self.config.protected_names,
            accurate_mem=False,
            ilp_solver=self.ilp_solver,
        ):

            md = model_ilp(
                hg,
                peak_budget=peak_budget,
                save_budget=max(save_budget),
                ilp_solver_params=ilp_solver_params,
                accurate_mem=accurate_mem,
                protected_names=protected_names,
                activation_offload=self.config.activation_offload,
                minor_offload_size = self.config.minor_offload_size,
                **self.config.model_kwargs
            )
            md.build()
            # print(f"model building: {time.time()-start}")
            sols = set()
            for sv_budget in np.sort(save_budget)[::-1]:
                md.add_abar_constraint(sv_budget)
                md.solve(ilp_solver)
                # if not md.feasible:
                # if print_result:
                # print("Not feasible solution")
                # return []
                if md.feasible == 1:
                    if print_result:
                        # if True:
                        # print(
                        #     f"Solution with obj: {md.md.getObjective().getValue()}"
                        # )
                        print(
                            f"Solve Hgraph {hg.name} with {len(hg.list_HCNs)} nodes takes {md.solve_time:03f}s, used {psutil.virtual_memory().used}B memory"
                        )
                    loss_idx = md.loss_idx
                    # if solver_name[0] == "gurobi":
                    #     time_mem = (
                    #         md.md.getObjective().getValue(),  # time
                    #         md.U[(loss_idx, loss_idx)].getValue(),  # save_mem
                    #     )
                    # else:
                    time_mem = (
                        md.md.objective.value(),
                        md.U[(loss_idx, loss_idx)].value(),  # save_mem
                    )
                    if not time_mem in sols:
                        # start = time.time()

                        sols.add(time_mem)
                        op_sched = schedule(md)
                        # if md.md.status==2:
                        #     status = "opt"
                        # elif md.md.status==9:
                        #     status = "early_stp"
                        # else:
                        #     status = md.md.status

                        op_sched.solver = f"HILP_{md.status}"
                        list_op_sched.append(op_sched)
                        # print(f"scheduling: {time.time()-start}")
                else:  # if infeasible, no need to try smaller budget
                    return list_op_sched, md
            # if accurate_mem:print(md.feasible)
            return list_op_sched, md if accurate_mem else None

        list_op_sched, md = solve_md(
            ilp_solver_params=ilp_solver_params,
            model_ilp=self.model_ilp,
            accurate_mem=self.config.accurate_mem,
            protected_names=self.config.protected_names,
            ilp_solver=self.ilp_solver,
        )
        if md is not None:
            self.md = md
            self.solving_time = None if not md.feasible else md.solving_time
            self.status = md.status
        return list_op_sched

    def get_activation_offload(self, op_sched):
        self.bandwidth = 1e7
        """
        With all the torch.cuda.Event and stream.wait_event() to synchornize
        between offload/prefetch, in the order of Main>Ofl>Del during fwd and Prf>Main during bwd

        """
        cluster = op_sched.cluster
        op_list = op_sched.op_list
        fwd_op_list = op_list[: op_sched.loss_idx]
        bwd_op_list = op_list[op_sched.loss_idx :]
        prefetch_op_list = []
        phantoms = op_sched.phantoms

        # for op in op_sched.op_list:
        #     if isinstance(op, ComputeOp):

        offload_ops = {}
        prefetch_ops = {}

        for anode in phantoms:
            if not anode.allocation_type == "data":
                continue
            i = max(
                min(op_sched.occurrences[ComputeOp(cnode).name]) for cnode in anode.deps
            )
            op_sched.op_list[i].record_event = True
            comp_op = op_sched.op_list[i]
            offload_op = OffloadOp(Activation(anode), time=anode.mem / self.bandwidth)
            offload_op.wait_events.append([comp_op.op_type, comp_op.target.name])
            offload_op.record_event = True
            # fwd_op_list.append(offload_op)
            offload_ops[offload_op] = i


            del_op = DeleteOp(Activation(anode))
            del_op.wait_events.append([offload_op.op_type, offload_op.target.name])
            users = [min(op_sched.occurrences[ComputeOp(cnode).name]) 
                    for cnode in anode.users_real
                    if cnode.is_fwd]
            if users:
                user_op = op_sched.op_list[max(users)]
                user_op.record_event = True
                del_op.wait_events.append([user_op.op_type, user_op.target.name])
            # fwd_op_list.append(del_op)
            offload_ops[del_op] = i

            # prefetch_op_list.append()
            alloc_op = AllocateOp(Activation(anode))
            prefetch_op = PrefetchOp(Activation(anode), time=anode.mem / self.bandwidth)
            prefetch_op.record_event = True
            prefetch_op.wait_events.append([offload_op.op_type, offload_op.target.name])
            # prefetch_op_list.append(prefetch_op)
            i = 0
            for j, op in enumerate(bwd_op_list):
                if isinstance(op, ComputeOp) and op.target in anode.users_real.union(anode.users_fake):
                    op.wait_events.append([prefetch_op.op_type, prefetch_op.target.name])
                    i = j

            prefetch_ops[alloc_op] = j
            prefetch_ops[prefetch_op] = j

        def simulate_schedule(comp_ops, prf_ops, ofl_ops):
            op_time = {}
            time = 0
            for op in prf_ops:
                op_time[op.name] = time
                time += op.time

            time = 0
            for op in comp_ops:
                wait_time = [op_time[f"{e[0]}({e[1]})"] 
                            for e in op.wait_events 
                            if f"{e[0]}({e[1]})" in op_time]
                if wait_time:
                    time = max(wait_time+[time])
                time += op.time
                op_time[op.name] = time
            
            time = 0
            for op in ofl_ops:
                wait_time = [op_time[f"{e[0]}({e[1]})"] 
                            for e in op.wait_events 
                            if f"{e[0]}({e[1]})" in op_time]
                if wait_time:
                    time = max(wait_time+[time])
                time += op.time
                op_time[op.name] = time
            return op_time

        fwd_op_time = simulate_schedule(fwd_op_list, [], offload_ops)
        bwd_op_time = simulate_schedule(bwd_op_list, prefetch_ops, [])

        fwd_op_list += list(sorted(offload_ops.keys(), key=lambda x:offload_ops[x]))
        bwd_op_list = bwd_op_list[:1] + list(sorted(prefetch_ops.keys(), key=lambda x:prefetch_ops[x])) + bwd_op_list[1:]

        new_op_sched = OpSchedule(
            fwd_op_list + bwd_op_list,
            loss_idx=len(fwd_op_list),
            cluster=cluster,
            init_op_list=op_sched.init_op_list,
        )

        fwd_wait_time = max(fwd_op_time.values()) - new_op_sched.fwd_time
        bwd_wait_time = max(bwd_op_time.values()) - new_op_sched.bwd_time
        
        new_op_sched.fwd_wait_time = fwd_wait_time
        new_op_sched.bwd_wait_time = bwd_wait_time

        return new_op_sched

    def add_offload_sched_hcn(self, hcn):
        scheds = [sched for sched in hcn.list_sched 
                    if sched.offload_mem == 0 
                    and sched.prefetch_mem == 0 ]
        for sched in scheds:
            hcn.list_sched.append(self.get_activation_offload(sched))



# ==========================
# definition file of RK_Sequence
# based on rotor/algorithms/sequence.py
# ==========================

# from .def_op import OpSchedule
from ...op_schedule import Op
from typing import List
from copy import deepcopy

ref_print_atoms = [True]

# ==========================
#  ====== Seq Atom Op =======
# ==========================
class SeqAtomOp:
    def __init__(self, op):
        header = f"{op.op_type} {op.main_target}"
        self.name = header
        self.time = op.time
        self.op = op

    def __str__(self):
        return self.name


# ==========================


# ==========================
# ========= Seq Op =========
# ==========================
class SeqOp:
    pass


class SeqLoss(SeqOp):
    def __init__(self):
        self.time = 0
        self.mem = 0

    def __str__(self):
        return "Loss"


# ** Seq Block Op **
# -> Subclasses : Fn,Fc,Fe,Bwd
class SeqBlockOp(SeqOp):
    def __init__(self, name, index, op_list: List[Op]):
        self.name = name
        self.index = index
        self.op_list = op_list
        # self.time = self.op_sched.time  # sum(o.time for o in body)
        # self.mem = self.op_sched.save_mem[-1]  # sum(o.mem  for o in body)
        # self.overhead = self.op_sched.overhead

    def __str__(self):
        return f"{self.name}_{self.index}"
        # header = f"{self.name} Block {self.index} in {self.time}"
        # if ref_print_atoms[0]:
        #     sb = "\n|  ".join([o.__str__() for o in self.op_sched.op_list])
        #     return f"{header}\n{sb}"
        # else:
        #     return header


class SeqBlockFn(SeqBlockOp):
    def __init__(self, *args):
        super().__init__("Fn", *args)


class SeqBlockFc(SeqBlockOp):
    def __init__(self, *args):
        super().__init__("Fc", *args)


class SeqBlockFe(SeqBlockOp):
    def __init__(self, index, option, op_list):
        super().__init__("Fe", index, op_list)
        self.option = option
    def __str__(self):
        return f"{self.name}_{self.index}_{self.option}"

class SeqBlockBwd(SeqBlockOp):
    def __init__(self, *args):
        super().__init__("Bwd", *args)


# ==========================


# ==========================
# ======== Sequence ========
# ==========================


class RK_Sequence:
    def __init__(self, l=None):
        if l:
            self.seq = l
        else:
            self.seq = []

    def __str__(self):
        return "\n".join([str(o) for o in self.seq])

    def insert(self, op: SeqOp):
        self.seq.append(op)

    def insert_seq(self, sub_seq):
        self.seq.extend(sub_seq.seq)

    def compute_time(self):
        return sum([o.time for o in self.seq])

    def cut_fwd_bwd(self):
        ln = len(self.seq)
        for i in range(ln):
            if isinstance(self.seq[i], SeqLoss):
                return (
                    RK_Sequence(list(self.seq[:i])),
                    RK_Sequence(list(self.seq[(i + 1) :])),
                )
        raise Exception("Can't cut a Sequence which doesn't have SeqLoss")

    def get_op_list(self):
        return [ deepcopy(op) for rk_op in self.seq
                    for op in rk_op.op_list ]

    def simulate(self, chain, display=True, stopAtLoss=False):
        return simulate(chain, self, display, stopAtLoss)

    def get_peak_memory(self, chain):
        return self.simulate(self, chain, display=False, stopAtLoss=False)
    def get_loss_memory(self, chain):
        return self.simulate(self, chain, display=False, stopAtLoss=True)

        
# ==========================
## Code to estimate the makespan and memory usage of a given sequence

from enum import Enum

Data = Enum("Data", ["X", "Xbar", "Y"])
def make_x(idx):
    return (Data.X, idx)
def make_xb(idx, option):
    return (Data.Xbar, idx, option)
def make_y(idx):
    return (Data.Y, idx)
def make_x_and_xb(index):
    return make_x(index), make_xb(index)

class Storage:
    def __init__(self, chain):
        self.chain = chain
        self.xs = {}
        self.xbs = {}
        self.ys = {}

    def add_x(self, idx):
        self.xs[idx] = True
    def add_xb(self, idx, option):
        assert not self.has_xb(idx)
        self.xbs[idx] = option
    def add_y(self, idx):
        self.ys[idx] = True

    def remove_x(self, idx):
        del self.xs[idx]
    def remove_xb(self, idx):
        del self.xbs[idx]
    def remove_y(self, idx):
        del self.ys[idx]

    def remove_x_or_xb(self, idx):
        if self.has_x(idx):
            self.remove_x(idx)
        else:
            self.remove_xb(idx)

    def has_x(self, idx):
        return idx in self.xs
    def has_y(self, idx):
        return idx in self.ys
    def has_xb(self, idx):
        return idx in self.xbs

    def get_xb_option(self, idx):
        try: 
            return self.xbs[idx]
        except IndexError:
            return None
 
    def usage(self):
        result =  sum(self.chain.cw[i] for i in self.xs)
        result += sum(self.chain.cw[i] for i in self.ys)
        result += sum(self.chain.cbw[i][o] for i, o in self.xbs.items())
        return result

    def __str__(self):
        items = [(i, f"x{i}") for i in self.xs] + [(i, f"y{i}") for i in self.ys] + [(i, f"xb{i}_{o}") for i, o in self.xbs.items()]
        items.sort(key=lambda x: x[0])
        return ", ".join(name for _, name in items)

# def elementUsage(chain, e):
#     t, idx = e[0], e[1]
#     if t is Data.X or Data.Y:
#         return self.cw[idx]
#     else:
#         option = e[2]
#         return self.cbw[idx][option]

# def memUsage(self, storage):
#     return sum(self.elementUsage(e) for e in storage)

# Simulates the execution of the sequence
# Returns the maximum memory usage. If stopAtLoss, returns the memory usage at the time of computing Loss
def simulate(chain, sequence, display=True, stopAtLoss=False):
    if display:
        ref_print_atoms = [False]
    l = chain.ln
    mem = Storage(chain)
    mem.add_x(0)
    maxUsage = mem.usage()
    for op in sequence.seq:
        if display: 
            print("Before", op, f"Usage={mem.usage()}")
            print("Contents:", mem)
        opType = type(op)
        if opType is SeqLoss:
            if not mem.has_x(l) and not mem.has_xb(l):
                raise ValueError(f"Before {op}: no X{l} or Xb{l} in memory")
            mem.add_y(l)
            used = mem.usage()
            opUsage =  used + chain.bwd_tmp[l][0] ##Only one option for Loss
            if stopAtLoss:
                return used
        else:
            index = op.index
            if not mem.has_x(index) and not mem.has_xb(index):
                raise ValueError(f"Before {op}: no X{index} or Xb{index} in memory")
            if opType is SeqBlockFe:
                mem.add_xb(index+1, op.option)
                opUsage = mem.usage() + chain.fwd_tmp[index][op.option]
            if opType is SeqBlockFn:
                mem.add_x(index+1)
                opUsage = mem.usage() + chain.ff_fwd_tmp[index]
                mem.remove_x_or_xb(index)
            if opType is SeqBlockFc:
                mem.add_x(index+1)
                opUsage = mem.usage() + chain.ff_fwd_tmp[index]
            if opType is SeqBlockBwd:
                if not mem.has_y(index+1) or not mem.has_xb(index+1):
                    raise ValueError(f"Before {op}: no Y{index+1} or Xb{index+1} in memory")
                else:
                    mem.add_y(index)
                    option = mem.get_xb_option(index+1)
                    opUsage = mem.usage() + chain.bwd_tmp[index][option]
                    if mem.has_x(index):
                        mem.remove_x(index)
                    # Do not remove xb(index), it will be useful for BWD(index-1)
                    mem.remove_y(index+1)
                    mem.remove_xb(index+1)
        if display: print("During:", opUsage)
        maxUsage = max(maxUsage, opUsage)
    return maxUsage




from pulp import LpVariable

def set_hcn_list_sched(hcn, list_sched):
    """
    As part of preprocess for the hierarchical solver, we selection
    schedules and store in hcn.list_sched, which is temporarily used 
    by the ILP. 
    """
    setattr(hcn, "list_sched", list_sched)

def clean_hcn_list_sched(hcn):
    delattr(hcn, "list_sched")

def set_hg_parameter_groups(hg, parameter_groups):
    setattr(hg, "parameter_groups", parameter_groups)

def clean_hg_parameter_groups(hg):
    delattr(hg, "parameter_groups")



class RkLpVariable(LpVariable):
    def __init__(self, name, lowBound=None, upBound=None, cat="Continuous", e=None, solution=None):
        super().__init__(name=name, lowBound=lowBound, upBound=upBound, cat=cat, e=e)
        self.solution = solution

    def value(self):
        return self.solution or self.varValue

    @classmethod
    def dicts(
        cls,
        name,
        indices=None,
        lowBound=0,
        upBound=1,
        cat="Continuous",
        indexStart=[],
    ):
        d = {}
        for index in indices:
            var_name = name + "_" + "_".join([str(i) for i in index])
            d[index] = RkLpVariable(
                var_name, lowBound=lowBound, upBound=upBound, cat=cat
            )
        return d

    def __repr__(self):
        if self.varValue:return str(self.varValue)
        return super().__repr__()

    def prefill(self, value):
        self.setInitialValue(value)
        self.fixValue()



import rkgb
import torch
import numpy as np
from copy import deepcopy

from rkgb.lowlevel.ast_add_on import ast_to_str
from rkgb.lowlevel.measure import TimerCPU
from rkgb.core.hierarchical import HierarchicalGraph, HierarchicalCluster
from rkgb.core.backward import ComputationNode
from rkgb.lowlevel.constants import init_target_string

from .main import Solver
from ..op_schedule import (
    OpSchedule,
    ComputeOp,
    DeleteOp,
    Activation,
    OffloadOp,
    PrefetchOp,
    AllocateOp,
)
import time
import psutil
import dataclasses

class CheapSolver(Solver):
    """
    Cheap solver that solves a cluster with the following heuristic:
    * all the cheap operations are recomputed and
    * all the phantoms are offloaded during forward and prefetched during backward.

    An operation is cheap if the ratio of computation time to offload time is below the cheap_factor
    """

    @dataclasses.dataclass
    class Config:
        bandwidth: float = 1e7
        add_offload: bool = False
        cheap_factor: float = 2.0

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def solve(self, cluster: HierarchicalCluster, budget=None):
        avg_time = np.mean(
            [cnode.time for cnode in cluster.list_cnodes if cnode.time is not None]
        )

        def is_cheap(cnode):
            if not cnode.is_fwd:
                return False
            if cnode.time is None:
                return False
            # return cnode.time < avg_time
            return cnode.time < sum(anode.mem for anode in cnode.users)/self.config.bandwidth/self.config.cheap_factor

        cnode_idx = {cnode.name: i for i, cnode in enumerate(cluster.list_cnodes)}
        loss_idx = len([cnode for cnode in cluster.list_cnodes if cnode.is_fwd]) - 1
        cheap_cnodes = {}
        for cnode in cluster.list_cnodes:
            user_cnodes = [
                user_cnode
                for anode in cnode.users
                for user_cnode in anode.users_real
                if user_cnode.name in cnode_idx
            ]
            # has_bwd_user = any(not user_cnode.is_fwd for user_cnode in user_cnodes)
            if is_cheap(cnode):  # and has_bwd_user:
                cheap_cnodes[cnode.name] = cnode

        anodes_del_idx = {i: [] for i, _ in enumerate(cluster.list_cnodes)}
        output_anodes = []
        for anode in cluster.list_anodes:
            if "source" in anode.name or anode in cluster.interfaces["output_data_anodes"]:
                continue
            user_indices = []
            for cnode in anode.users_real:
                if cnode.name in cnode_idx:
                    user_indices.append(cnode_idx[cnode.name])
                if cnode.name in cheap_cnodes:
                    user_indices.append(loss_idx)

            if user_indices:
                last_user_idx = max(user_indices)
                anodes_del_idx[last_user_idx].append(anode)
            else:
                output_anodes.append(anode)

            regenerated = any(cnode.name in cheap_cnodes for cnode in anode.deps)
            if regenerated and all(cnode.is_fwd for cnode in anode.deps):
                anodes_del_idx[loss_idx].append(anode)

            if (
                all(is_cheap(cnode) for cnode in anode.deps)
                and anode.allocation_type == "data"
            ):
                user_indices_fwd = [
                    cnode_idx[cnode.name]
                    for cnode in anode.users_real
                    if cnode.is_fwd and cnode.name in cnode_idx
                ]
                if user_indices_fwd:
                    last_user_idx_fwd = max(user_indices_fwd)
                    anodes_del_idx[last_user_idx_fwd].append(anode)

        fwd_op_list = []
        bwd_op_list = []
        for i, cnode in enumerate(cluster.list_cnodes[:loss_idx]):
            fwd_op_list.append(ComputeOp(cnode, disabled="loss" in cnode.name))
            # if "loss" in cnode.name:
            #     loss_idx = i
            #     break
            for anode in anodes_del_idx[i]:
                fwd_op_list.append(DeleteOp(Activation(anode)))

        fwd_op_list.append(ComputeOp(cluster.list_cnodes[loss_idx], disabled=True))

        for anode in output_anodes:
            bwd_op_list.append(DeleteOp(Activation(anode)))
        for cnode in cheap_cnodes.values():
            bwd_op_list.append(ComputeOp(cnode))
        for anode in anodes_del_idx[loss_idx]:
            bwd_op_list.append(DeleteOp(Activation(anode)))

        for i, cnode in enumerate(cluster.list_cnodes[loss_idx + 1 :]):
            bwd_op_list.append(ComputeOp(cnode))
            for anode in anodes_del_idx[i + loss_idx + 1]:
                bwd_op_list.append(DeleteOp(Activation(anode)))

        op_sched = OpSchedule(
            fwd_op_list + bwd_op_list, loss_idx=len(fwd_op_list) - 1, cluster=cluster
        )
        if self.config.add_offload:
            op_sched = self.add_activation_offload(op_sched)
        return [op_sched]


    def add_activation_offload(self, op_sched: OpSchedule) -> OpSchedule:
        """
        With all the torch.cuda.Event and stream.wait_event() to synchornize
        between offload/prefetch, in the order of Main>Ofl>Del during fwd and Prf>Main during bwd

        """
        cluster = op_sched.cluster
        op_list = op_sched.op_list
        fwd_op_list = op_list[: op_sched.loss_idx]
        bwd_op_list = op_list[op_sched.loss_idx :]
        prefetch_op_list = []
        phantoms = op_sched.phantoms

        # for op in op_sched.op_list:
        #     if isinstance(op, ComputeOp):

        offload_ops = {}
        prefetch_ops = {}

        for anode in phantoms:
            if not anode.allocation_type == "data":
                continue
            i = max(
                min(op_sched.occurrences[ComputeOp(cnode).name]) for cnode in anode.deps
            )
            op_sched.op_list[i].record_event = True
            comp_op = op_sched.op_list[i]
            offload_op = OffloadOp(Activation(anode), time=anode.mem / self.config.bandwidth)
            offload_op.wait_events.append([comp_op.op_type, comp_op.target.name])
            offload_op.record_event = True
            # fwd_op_list.append(offload_op)
            offload_ops[offload_op] = i


            del_op = DeleteOp(Activation(anode))
            del_op.wait_events.append([offload_op.op_type, offload_op.target.name])
            users = [min(op_sched.occurrences[ComputeOp(cnode).name]) 
                    for cnode in anode.users_real
                    if cnode.is_fwd]
            if users:
                user_op = op_sched.op_list[max(users)]
                user_op.record_event = True
                del_op.wait_events.append([user_op.op_type, user_op.target.name])
            # fwd_op_list.append(del_op)
            offload_ops[del_op] = i

            # prefetch_op_list.append()
            alloc_op = AllocateOp(Activation(anode))
            prefetch_op = PrefetchOp(Activation(anode), time=anode.mem / self.config.bandwidth)
            prefetch_op.record_event = True
            prefetch_op.wait_events.append([offload_op.op_type, offload_op.target.name])
            # prefetch_op_list.append(prefetch_op)
            i = 0
            for j, op in enumerate(bwd_op_list):
                if isinstance(op, ComputeOp) and op.target in anode.users_real.union(anode.users_fake):
                    op.wait_events.append([prefetch_op.op_type, prefetch_op.target.name])
                    i = j

            prefetch_ops[alloc_op] = j
            prefetch_ops[prefetch_op] = j

        def simulate_schedule(comp_ops, prf_ops, ofl_ops):
            op_time = {}
            time = 0
            for op in prf_ops:
                op_time[op.name] = time
                time += op.time

            time = 0
            for op in comp_ops:
                wait_time = [op_time[f"{e[0]}({e[1]})"] 
                            for e in op.wait_events 
                            if f"{e[0]}({e[1]})" in op_time]
                if wait_time:
                    time = max(wait_time+[time])
                time += op.time
                op_time[op.name] = time
            
            time = 0
            for op in ofl_ops:
                wait_time = [op_time[f"{e[0]}({e[1]})"] 
                            for e in op.wait_events 
                            if f"{e[0]}({e[1]})" in op_time]
                if wait_time:
                    time = max(wait_time+[time])
                time += op.time
                op_time[op.name] = time
            return op_time

        fwd_op_time = simulate_schedule(fwd_op_list, [], offload_ops)
        bwd_op_time = simulate_schedule(bwd_op_list, prefetch_ops, [])

        fwd_op_list += list(sorted(offload_ops.keys(), key=lambda x:offload_ops[x]))
        bwd_op_list = bwd_op_list[:1] + list(sorted(prefetch_ops.keys(), key=lambda x:prefetch_ops[x])) + bwd_op_list[1:]

        new_op_sched = OpSchedule(
            fwd_op_list + bwd_op_list,
            loss_idx=len(fwd_op_list),
            cluster=cluster,
            init_op_list=op_sched.init_op_list,
        )

        fwd_wait_time = max(fwd_op_time.values()) - new_op_sched.fwd_time
        bwd_wait_time = max(bwd_op_time.values()) - new_op_sched.bwd_time
        
        new_op_sched.fwd_wait_time = fwd_wait_time
        new_op_sched.bwd_wait_time = bwd_wait_time

        return new_op_sched



import time
import warnings
from rkgb.core.hierarchical import HierarchicalGraph, HierarchicalCluster
from rkgb.core.backward import ComputationNode
from dataclasses import dataclass

from ..main import Solver, get_cluster_budget
from .def_chain import RK_Chain
from .rotor_solver import seq_builder, solve_dp_functional
from ...op_schedule import OpSchedule, ComputeOp

class RK_rotor(Solver):
    '''rk_rotor solver, adapted to the rockmate framework from the Rotor dynamic-programming based algorithm

    The rotor algorithm is originally implemented in https://gitlab.inria.fr/hiepacs/rotor and described
    in the paper:

    Olivier Beaumont, Lionel Eyraud-Dubois, Julien Herrmann, Alexis Joly, Alena Shilova. Optimal
    checkpointing for heterogeneous chains: how to train deep neural networks with limited memory. ACM
    Transactions on Mathematical Software, In press. https://hal.science/hal-02352969v2

    This version is modified to allow choosing between several rematerialization options for each layer.
    '''
    @dataclass
    class Config:
        mem_unit: int = 1024**2
        force_python: bool = False

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def is_sequential(self, hg: HierarchicalGraph):
        loss_idx = hg.list_HCNs.index(hg.loss_hcn)
        for i, hcn in enumerate(hg.list_HCNs[:loss_idx]):  # toposorted
            users = set(hcn_user for han in hcn.users 
                        for hcn_user in han.users
                        if hcn_user.is_fwd)
            if len(users)>1:
                return False
        input_data = hg.input_data_HANs
        users_of_input = set(hcn_user for han in hg.input_data_HANs
                             for hcn_user in han.users
                             if hcn_user.is_fwd)
        if len(users_of_input) > 1:
            return False
        return True

    def solve(
        self,
        cluster,
        budgets=None,
    ):
        if budgets is None:
            budgets = get_cluster_budget(cluster.representee_cluster)
        if isinstance(cluster, HierarchicalCluster):
            list_seq = []
            for hg in cluster.partitionings:
                list_seq.extend(self.solve_hg(hg, budgets))
            return list_seq
        else:
            warnings.warn(f"Unrecognized input type {type(cluster)}")

    def solve_hg(self, hg: HierarchicalGraph, budgets=[]):
        if not self.is_sequential(hg):
            return []
        else:
            chain = RK_Chain(hg, self.config.mem_unit)
            opt_table = self.solve_rk_chain(chain, max(budgets))

            list_op_sched = []
            for budget in budgets:

                try:
                    seq = seq_builder(
                        chain,
                        self.discretize_budget(chain, budget),
                        opt_table
                    )
                except ValueError as e:
                    if not "budget" in str(e):  # not enough budget
                        raise e
                    else:
                        continue

                fwd_seq, bwd_seq = seq.cut_fwd_bwd()
                
                fwd_op_list = fwd_seq.get_op_list()
                op_sched = OpSchedule(
                    fwd_op_list + [ComputeOp(ComputationNode("loss"))] + bwd_seq.get_op_list(),
                    ## TODO: why not -1 here ???
                    loss_idx = len(fwd_op_list),
                    cluster=hg.cluster,
                )
                list_op_sched.append(op_sched)
            return list_op_sched

    def discretize_budget(self, chain, budget):
        ## Both input and output stay in memory all along
        ##   (this is a Rotor assumption, because Rotor can not remove them from memory)
        ##   (and even Rockmate can not assume too much, because the user has access to them)
        return int(budget // self.config.mem_unit) - chain.cw[0] - chain.cw[chain.ln]

    # Returns the opt_table
    def solve_rk_chain(self, chain, budget):
        start = time.time()

        mmax = self.discretize_budget(chain, budget)
        opt_table = solve_dp_functional(chain, mmax, force_python=self.config.force_python)

        end = time.time()
        self.DP_solve_time = end - start
        return opt_table



# The ILP algorithm is inspired from Checkmate: https://github.com/mllg/checkmate
from typing import Dict, Any
import time
import numpy as np
from copy import deepcopy
from pulp import (
    LpVariable,
    LpProblem,
    LpMinimize,
    lpSum,
    get_solver,
    listSolvers,
    LpStatus,
)

from rkgb.core.hierarchical import HierarchicalGraph
from .ilp_utils import RkLpVariable


class ModelPULP:
    """
    Build the ILP model by given Hgraph and budget.
    RN this model will take a rk_chain to solve the solution.
    """

    def __init__(
        self,
        hgraph: HierarchicalGraph,
        peak_budget: int,
        save_budget=None,
        ilp_solver_params: Dict[str, Any] = {
            "LogToConsole": 0,
            "IntegralityFocus": 1,
            "TimeLimit": 4 * 60,
        },
        gcd=None,
        accurate_mem=False,
        protected_names=[],
        **kwargs,
    ):
        self.gcd = gcd if gcd else 1024**2
        self.peak_budget = peak_budget / self.gcd
        if save_budget:
            self.save_budget = save_budget / self.gcd
        else:
            self.save_budget = peak_budget / self.gcd

        self.ilp_solver_params = ilp_solver_params
        self.feasible = None
        self.solve_time = None
        self.single_fwd = False
        self.single_bwd = False
        self.protected_names = protected_names
        self.hgraph = hgraph
        self.md = LpProblem(f"rockmateMILP", LpMinimize)
        self.use_correction_term = accurate_mem

        self.config_ilp()

    def build(self):
        # OVERWRITTING METHOD IN OFFLOAD
        self.add_variables()
        self.add_constraints()
        self.add_objective()

    def config_ilp(self):
        self.hcn2sub_c = []
        self.list_list_sched = []
        self.sub_clusters = []
        self.nSched = []  # number of fwd-bwd schedule
        self.nComp = []  # number to run compute, =nSched if bwd, =nSched + 1 if fwd
        self.time = []
        self.overhead = []
        self.bin_type = "Binary"

        for i, hcn in enumerate(self.hgraph.list_HCNs):
            if "Loss" in hcn.name:
                self.loss_idx = i
            if hcn.sub_cluster is None:
                # only when hcn is fwd with requires_grad=False
                self.hcn2sub_c.append(None)
                self.nComp.append(1)
                self.nSched.append(0)
                self.time.append([hcn.ff_time])
                self.overhead.append([hcn.ff_overhead / self.gcd])
            else:
                ff = True#not hcn.sub_cluster.is_bottom
                if hcn.is_fwd:
                    self.list_list_sched.append(hcn.list_sched)
                    self.sub_clusters.append(hcn.sub_cluster)
                j = self.sub_clusters.index(hcn.sub_cluster)
                self.hcn2sub_c.append(j)
                list_sched = self.list_list_sched[j]  # hcn bwd does not have list_sched
                # self.hcn2sub_c.append(len(self.list_list_sched) - 1)
                self.nComp.append(len(list_sched) + (1 if hcn.is_fwd and ff else 0))
                self.nSched.append(len(list_sched))

                if hcn.is_fwd:
                    # add fast forward to the options (final one)
                    self.time.append(
                        [op_sched.fwd_time for op_sched in list_sched] + 
                        ([hcn.ff_time] if ff else [])
                    )
                    self.overhead.append(
                        [op_sched.fwd_overhead / self.gcd for op_sched in list_sched]
                        + ([hcn.ff_overhead / self.gcd] if ff else [])
                    )
                else:
                    self.time.append([op_sched.bwd_time for op_sched in list_sched])
                    self.overhead.append(
                        [op_sched.bwd_overhead / self.gcd for op_sched in list_sched]
                    )
        self.sub_c2hcn = [
            [] for _ in self.list_list_sched
        ]  # index of sub_cluster to index of hcn
        for i, j in enumerate(self.hcn2sub_c):
            if j is None:
                continue
            self.sub_c2hcn[j].append(i)
        self.mem = [han.mem / self.gcd for han in self.hgraph.list_HANs]
        self.saved_mem = [
            [op_sched.mem / self.gcd for op_sched in list_sched]
            for list_sched in self.list_list_sched
        ]

        self.T = len(self.hgraph.list_HCNs)
        self.I = len(self.hgraph.list_HANs)
        self.J = len(self.list_list_sched)

        self.hcn2idx = {hcn: i for i, hcn in enumerate(self.hgraph.list_HCNs)}
        self.han2idx = {hcn: i for i, hcn in enumerate(self.hgraph.list_HANs)}

        self.protected_indices = []
        self.input_grad_indices = []
        self.input_data_indices = []
        self.han_deps = []
        self.han_users = []
        self.han_users_real = []  # without sched-dependent users

        for i, han in enumerate(self.hgraph.list_HANs):
            if han.anode.name in self.protected_names:
                self.protected_indices.append(i)
            if han in self.hgraph.input_grad_HANs:
                self.input_grad_indices.append(i)
            if han in self.hgraph.input_data_HANs:
                self.input_data_indices.append(i)
            self.han_deps.append([self.hcn2idx[hcn] for hcn in han.deps])
            self.han_users.append(
                [self.hcn2idx[hcn] for hcn in han.users if hcn in self.hcn2idx]
            )  # current sub-graph users

        self.han_users_real = self.han_users.copy()
        self.hcn_users = [
            [self.han2idx[han] for han in self.hgraph.list_HCNs[k].users]
            for k in range(self.T)
        ]  # outputs of hcn

        #### Update edges based on .dep_interfaces_data
        #### In certain schedules, BWD depends on input/output data
        for k, hcn in enumerate(self.hgraph.list_HCNs):
            if hcn.sub_cluster is None:
                continue
            list_sched = self.list_list_sched[self.hcn2sub_c[k]]
            cluster = self.sub_clusters[self.hcn2sub_c[k]]
            for op_sched in list_sched:
                # for i_ in op_sched.dep_interfaces_data:
                for anode in op_sched.dep_interfaces_data:
                    self_anode = cluster.translate_representee_node(anode)
                    # Without specifying schedule, we assume it's possible to use han here
                    for i in range(self.I):
                        if (
                            self_anode.name == self.hgraph.list_HANs[i].anode.name
                            and k not in self.han_users[i]
                        ):
                            self.han_users[i].append(k)

        ##############################

        self.create_list = [(k, i) for k in range(self.T) for i in self.hcn_users[k]]
        self.delete_list = [
            (k, i) for i in range(self.I) for k in self.han_deps[i] + self.han_users[i]
        ]

        self.Cr = len(self.create_list)
        self.De = len(self.delete_list)
        self.W = len(self.sub_clusters)

    def krange(self, t):
        if self.single_fwd:
            return [t]
        elif self.single_bwd and t > self.loss_idx:
            return list(range(self.loss_idx)) + [t]
        return list(range(t + 1))

    def next_idx(self, t, i, upper_triangle=False):
        # if upper_triangle, consider the case when i>t
        if t == self.T - 1:
            if i < max(self.krange(t)):
                t_ = t
                i_ = self.krange(t)[self.krange(t).index(i) + 1]
            else:
                t_ = 0
                i_ = 0
        else:
            end = max(self.krange(t))
            if i < end:
                t_ = t
                i_ = self.krange(t)[self.krange(t).index(i) + 1]
            else:
                i_ = min(self.krange(t + 1))
                t_ = t + 1
        return (t_, i_)

    def add_objective(self, bandwidth_cost=0.01):
        self.md += lpSum(self.Time[t, k] for t in range(self.T) for k in self.krange(t))

    def add_constraints(self):
        self.add_valid_constraints()
        self.add_memory_constrains()

        ##### Time constraints
        for t in range(self.T):
            for k in self.krange(t):
                self.md += self.Time[t, k] >= lpSum(
                    self.Comp[t, k, o] * self.time[k][o] for o in range(self.nComp[k])
                )

    def add_variables(self):
        self.Comp = RkLpVariable.dicts(
            f"Comp",
            [
                (t, k, o)
                for t in range(self.T)
                for k in self.krange(t)
                for o in range(self.nComp[k])
            ],
            cat=self.bin_type,
        )

        self.sumComp = {}
        for t in range(self.T):
            for k in self.krange(t):
                self.sumComp[t, k] = lpSum(
                    self.Comp[t, k, o] for o in range(self.nComp[k])
                )

        for t in range(self.T):
            for k in range(self.T):
                if k not in self.krange(t):
                    for o in range(self.nComp[k]):
                        self.Comp[t, k, o] = 0
                    self.sumComp[t, k] = 0

        self.AliveP = RkLpVariable.dicts(
            f"AliveP",
            [
                (t, j, o)
                for t in range(self.T + 1)
                for j, list_sched in enumerate(self.list_list_sched)
                for o in range(len(list_sched))
                # if t-1 in self.sub_c2hcn[j]
            ],
            cat=self.bin_type,
        )

        self.sumAliveP = {}
        for j in range(self.J):
            for t in range(self.T + 1):
                self.sumAliveP[t, j] = lpSum(
                    self.AliveP[t, j, o] for o in range(len(self.list_list_sched[j]))
                )

        self.active_stages = dict()
        for i in range(self.I):
            self.active_stages[i] = []
            for t in range(self.T):
                for k in self.krange(t):
                    if k in self.han_deps[i] + self.han_users[i]:
                        self.active_stages[i].append(t)
            if not self.hgraph.list_HANs[i].deps:  # src node
                self.active_stages[i].append(-1)

        # to present whether one saved tensor can be inherited from the last stage
        self.AliveA = RkLpVariable.dicts(
            "AliveA",
            [
                (t, c)
                for t in range(1, self.T)
                for c, (k, i) in enumerate(self.create_list)
                #    if t-1 in self.active_stages[i]
            ],
            cat=self.bin_type,
        )  # activation

        self.AliveT = RkLpVariable.dicts(
            "AliveT",
            [
                (t, i)
                for t in range(self.T)
                for i in range(self.I)
                #    if t-1 in self.active_stages[i]
            ],
            cat=self.bin_type,
        )  # tensor that can be shared by acts

        self.create = RkLpVariable.dicts(
            "create",
            [
                (t, i)
                for t in range(self.T)
                for i in range(self.Cr)
                if self.create_list[i][0] in self.krange(t)
            ],
            cat=self.bin_type,
        )
        self.delete = RkLpVariable.dicts(
            "delete",
            [
                (t, i)
                for t in range(self.T)
                for i in range(self.De)
                if self.delete_list[i][0] in self.krange(t)
            ],
            cat=self.bin_type,
        )

        self.Time = RkLpVariable.dicts(
            "Time",
            [(t, k) for t in range(self.T) for k in self.krange(t)],
            lowBound=None,
            upBound=None,
            cat="Continuous",
        )
        self.prefill_compute()
        # if self.with_offload:
        #     add_offload_variables(self)

    def add_valid_constraints(self):
        # In the last stage, every source edge of input_grad should be alive or executed
        for i in self.input_grad_indices:
            for j_, (k_, i_) in enumerate(self.create_list):
                if i_ == i:
                    self.md += (
                        self.AliveA[self.T - 1, j_] + self.sumComp[self.T - 1, k_] == 1
                    )

        # # In the first stage, assume input data is alive
        # for i in self.input_data_indices:
        #     for j_, (k_, i_) in enumerate(self.create_list):
        #         if i_ == i:
        #             self.md += (self.AliveA[0, j_] ==  1)

        for j in range(self.J):
            bwd_i = max(self.sub_c2hcn[j])
            # Forward start with no phantoms
            # self.md += (
            #     lpSum(
            #         (self.AliveP[0, j, o])  # - self.Comp[bwd_i][T - 1, o])
            #         for o in range(self.nSched[bwd_i])
            #     )
            #     == 0
            # )
            # in the end of bwd, del every phantoms
            self.md += (
                lpSum(
                    (self.AliveP[self.T, j, o])  # - self.Comp[bwd_i][T - 1, o])
                    for o in range(self.nSched[bwd_i])
                )
                == 0
            )

        # options don't conflict
        for t in range(self.T):
            for k in self.krange(t):
                self.md += self.sumComp[t, k] <= 1
        for t in range(self.T + 1):
            for j in range(self.J):
                self.md += (
                    self.sumAliveP[t, j] <= 1
                )  # assuming two copies of saved tensors won't be kept at the same time

        #### Option-free constraints: from rk-checkmate
        self.md += (
            lpSum(self.sumComp[t, t] for t in range(self.T)) == self.T
        )  # diagonal should be executed
        self.md += (
            lpSum(
                self.sumComp[t, self.loss_idx]
                for t in range(self.T)
                if self.loss_idx in self.krange(t)
            )
            == 1
        )  # loss should be executed exactly once

        for t in range(self.T):
            for j in range(self.Cr):
                self.md += (
                    self.AliveA[t, j] <= self.AliveT[t, self.create_list[j][1]]
                )  # one edge created, memory is occupied
        for t in range(self.T - 1):
            for j in range(self.Cr):
                src_i = self.create_list[j][0]
                src = self.sumComp[t, src_i] if src_i in self.krange(t) else 0
                self.md += self.AliveA[t + 1, j] <= self.AliveA[t, j] + src
        for t in range(self.T):
            for j, (k, i) in enumerate(self.create_list):
                for k_ in self.han_users[i]:
                    self.md += (
                        self.sumComp[t, k_] <= self.sumComp[t, k] + self.AliveA[t, j]
                    )

        #### Options-related constraints
        for t in range(self.T):
            for j in range(self.J):
                fwd_i = min(self.sub_c2hcn[j])
                bwd_i = max(self.sub_c2hcn[j])
                for o in range(self.nSched[fwd_i]):
                    self.md += (
                        self.AliveP[t + 1, j, o]
                        <= self.AliveP[t, j, o] + self.Comp[t, fwd_i, o]
                    )  # phantoms can only be generated by fwd
                    self.md += (
                        self.AliveP[t + 1, j, o]
                        >= self.AliveP[t, j, o]
                        - self.Comp[t, bwd_i, o]
                        + self.Comp[t, fwd_i, o]
                    )  # phantoms can only be deleted by bwd
                    self.md += (
                        self.Comp[t, bwd_i, o]
                        <= self.AliveP[t, j, o] + self.Comp[t, fwd_i, o]
                    )

                    list_sched = self.list_list_sched[j]
                    cluster = self.sub_clusters[j]
                    # for i in list_sched[o].dep_interfaces_data:
                    for anode in list_sched[o].dep_interfaces_data:
                        self_anode = cluster.translate_representee_node(anode)
                        hcn = self.hgraph.list_HCNs[bwd_i]
                        # Tensor req_i is required by BWD
                        req_i = [han.anode.name for han in self.hgraph.list_HANs].index(
                            self_anode.name
                        )
                        # req_i = self.han2idx[anode]
                        for j_, (k_, i_) in enumerate(self.create_list):
                            if i_ == req_i:
                                self.md += (
                                    self.Comp[t, bwd_i, o]
                                    <= self.sumComp[t, k_] + self.AliveA[t, j_]
                                )

    def add_activation_deletion(self):
        # we don't keep eyes on the alive status all the time
        # only the steps when changes can happen
        self.alive = {}
        for t in range(self.T):
            for eidx, (k, i) in enumerate(self.delete_list):
                self.alive[(t, k, i)] = self.AliveT[t, i]
                self.alive[(t, k, i)] += lpSum(
                    self.create[t, eidx_c]
                    for eidx_c, (k_, i_) in enumerate(self.create_list)
                    if i_ == i and k_ <= k
                )
                self.alive[(t, k, i)] -= lpSum(
                    self.delete[t, eidx_d]
                    for eidx_d, (k_, i_) in enumerate(self.delete_list)
                    if i_ == i and k_ <= k
                )
                self.md += self.alive[(t, k, i)] >= 0
                self.md += self.alive[(t, k, i)] <= 1
                if (k, i) in self.create_list:
                    didx = self.delete_list.index((k, i))
                    self.md += (
                        self.alive[(t, k, i)] + self.delete[t, didx]
                        >= self.sumComp[t, k]
                    )

            for eidx, (k, i) in enumerate(self.create_list):
                # if k not in self.krange(t):
                #     continue
                self.md += self.create[t, eidx] <= self.sumComp[t, k]
                # self.md += self.create[t, eidx] >= self.sumComp[t, k] - self.alive[(t,k,i)]
            for i in range(self.I):
                if t + 1 < self.T:
                    pass
                    self.md += (
                        self.AliveT[t + 1, i]
                        == self.alive[(t, max(self.han_deps[i] + self.han_users[i]), i)]
                    )
                elif i not in self.protected_indices:
                    # in the end of bwd, del every HDN
                    step = None
                    for k in self.krange(t):
                        if k in self.han_deps[i] + self.han_users[i]:
                            step = k
                    if step is not None:
                        self.md += self.alive[(t, step, i)] == 0
                    # elif i>9:
                    #     print(i)
                    else:
                        self.md += self.AliveT[self.T - 1, i] == 0

        def _num_hazards(t, i, k):
            if i in self.protected_indices:
                return _max_num_hazards(t, i, k)
            if t + 1 < self.T:
                return (
                    1
                    - self.sumComp[t, k]
                    + self.AliveT[t + 1, i]
                    + lpSum(self.sumComp[t, j] for j in self.han_users[i] if j > k)
                )
            return (
                1
                - self.sumComp[t, k]
                + lpSum(self.sumComp[t, j] for j in self.han_users[i] if j > k)
            )

        def _max_num_hazards(t, i, k):
            num_uses_after_k = sum(1 for j in self.han_users[i] if j > k)
            if t + 1 < self.T:
                return 2 + num_uses_after_k
            return 1 + num_uses_after_k

        # delete when not needed
        for t in range(self.T):
            for eidx, (k, i) in enumerate(self.delete_list):
                self.md += 1 - self.delete[t, eidx] <= _num_hazards(t, i, k)

        # don't delete if still needed
        for t in range(self.T):
            for eidx, (k, i) in enumerate(self.delete_list):
                self.md += _max_num_hazards(t, i, k) * (
                    1 - self.delete[t, eidx]
                ) >= _num_hazards(t, i, k)
                if i in self.protected_indices:
                    self.md += self.delete[t, eidx] == 0

    def add_activation_mem(self):
        self.U = {}
        for t in range(self.T):
            self.U[t, 0] = (
                lpSum(self.AliveT[t, i] * self.mem[i] for i in range(self.I))
                + lpSum(
                    self.create[t, eidx] * self.mem[i]
                    for eidx, (k_, i) in enumerate(self.create_list)
                    if k_ == 0 and k_ in self.krange(t)
                )
                - lpSum(
                    self.delete[t, eidx] * self.mem[i]
                    for eidx, (k_, i) in enumerate(self.delete_list)
                    if k_ == 0 and k_ in self.krange(t)
                )
                + lpSum(
                    self.AliveP[t, j, o] * save_mem
                    for j in range(self.J)
                    for o, save_mem in enumerate(self.saved_mem[j])
                )
                + lpSum(  # if the first fwd operation creates phantoms
                    self.Comp[t, 0, o] * self.saved_mem[self.hcn2sub_c[0]][o]
                    for o in range(self.nSched[0])
                )
            )

        for t in range(self.T):
            for k in range(1, self.T):
                j = self.hcn2sub_c[k]
                self.U[t, k] = (
                    self.U[(t, k - 1)]
                    + lpSum(
                        self.create[t, eidx] * self.mem[i]
                        for eidx, (k_, i) in enumerate(self.create_list)
                        if k_ == k and k_ in self.krange(t)
                    )
                    - lpSum(
                        self.delete[t, eidx] * self.mem[i]
                        for eidx, (k_, i) in enumerate(self.delete_list)
                        if k_ == k and k_ in self.krange(t)
                    )
                )
                # if k < self.loss_idx:
                if self.hgraph.list_HCNs[k].is_fwd:
                    self.U[t, k] += lpSum(
                        self.Comp[t, k, o] * self.saved_mem[j][o]
                        for o in range(self.nSched[k])
                    )
                else:
                    if j is None:
                        continue
                    fwd_i = min(self.sub_c2hcn[j])
                    self.U[t, k] += lpSum(
                        (
                            self.AliveP[t + 1, j, o]
                            - self.Comp[t, fwd_i, o]
                            - self.AliveP[t, j, o]
                        )
                        * self.saved_mem[j][o]
                        for o in range(self.nSched[k])
                    )

    def save_mem(self, t, k):
        # OVERWRITTING METHOD IN OFFLOAD
        # it represents the saving memory AFTER deletion operations at step (t,k)
        return self.U[t, k]

    def overhead_mem(self, t, k):
        # OVERWRITTING METHOD IN OFFLOAD
        return self.act_overhead(t, k)

    def act_overhead(self, t, k):
        j = self.hcn2sub_c[k]
        if self.use_correction_term and not j is None:
            return self.correction_term_overhead(t, k)
        overhead = lpSum(
            self.Comp[t, k, o] * self.overhead[k][o] for o in range(self.nComp[k])
        )
        overhead += lpSum(
            self.mem[i_] * self.delete[t, eidx_d]
            for eidx_d, (k_, i_) in enumerate(self.delete_list)
            if k == k_
        )
        return [overhead]

    def correction_term_overhead(self, t, k):
        j = self.hcn2sub_c[k]
        cluster = self.sub_clusters[j]
        hcn = self.hgraph.list_HCNs[k]
        overhead_terms = []
        for o, op_sched in enumerate(self.list_list_sched[j]):
            correction_terms = (
                op_sched.fwd_overhead_correction
                if hcn.is_fwd
                else op_sched.bwd_overhead_correction
            )
            if not correction_terms:
                overhead_terms.append(
                    self.Comp[t, k, o] * self.overhead[k][o]
                    + lpSum(
                        self.mem[i_] * self.delete[t, eidx_d]
                        for eidx_d, (k_, i_) in enumerate(self.delete_list)
                        if k == k_
                    )
                )
            for correction in correction_terms:
                correction_term = 0
                overhead = (correction["save"] + correction["overhead"]) - (
                    op_sched.mem if hcn.is_fwd else 0
                )
                for inter_position, inter_mem in correction.items():
                    if inter_position == "save" or inter_position == "overhead":
                        continue

                    i_ = [han.anode.name for han in self.hgraph.list_HANs].index(
                        # self.sub_clusters[j]
                        cluster.translate_representee_node(
                            op_sched.cluster.list_anodes[inter_position[0]]
                        ).name
                    )
                    if inter_position[1] == "always":
                        not_kept_alive = 1
                    elif not inter_position[1]:  # ending status
                        if (k, i_) in self.delete_list:
                            eidx = self.delete_list.index((k, i_))
                            not_kept_alive = self.delete[t, eidx]
                        else:  # when output_data is not deps, but we care about it
                            # eidx = self.delete_list.index((k, i_))
                            k_ = max([kk for kk in self.han_deps[i_] if kk < k])
                            not_kept_alive = self.alive[(t, k_, i_)]
                    else:  # start status
                        eidx = self.create_list.index((k, i_))
                        not_kept_alive = self.create[t, eidx]
                    correction_term += not_kept_alive * inter_mem

                overhead_terms.append(
                    self.Comp[t, k, o] * overhead / self.gcd
                    + correction_term / self.gcd
                    + lpSum(
                        self.mem[i_] * self.delete[t, eidx_d]
                        for eidx_d, (k_, i_) in enumerate(self.delete_list)
                        if k == k_
                    )
                )
        return overhead_terms

    def add_memory_constrains(self):
        self.add_activation_deletion()
        self.add_activation_mem()
        self.add_abar_constraint(self.save_budget * self.gcd)

        for t in range(self.T):
            for k in self.krange(t):
                self.md += self.save_mem(t, k) >= 0
                self.md += self.save_mem(t, k) <= (self.peak_budget)
                for overhead in self.overhead_mem(t, k):
                    self.md += self.save_mem(t, k) + overhead <= self.peak_budget

    def add_abar_constraint(self, save_budget):
        # self.save_budget = save_budget / self.gcd
        for k in range(self.T):
            self.md += self.U[(self.loss_idx, k)] <= save_budget / self.gcd

    def prefill_compute(self):
        self.active_stages = dict()
        for i in range(self.I):
            self.active_stages[i] = []
            for t in range(self.T):
                for k in self.krange(t):
                    if k in self.han_deps[i] + self.han_users[i]:
                        self.active_stages[i].append(t)
            if not self.hgraph.list_HANs[i].deps:  # src node
                self.active_stages[i].append(-1)

        for j, list_sched in enumerate(self.list_list_sched):
            for o in range(len(list_sched)):
                # if (0,j,o) not in self.AliveP:
                self.AliveP[0, j, o] = 0
                for t in range(1, self.T + 1):
                    # if (t,j,o) not in self.AliveP:
                    if not t - 1 in self.sub_c2hcn[j]:
                        self.AliveP[t, j, o] = self.AliveP[t - 1, j, o]

        for c, (k, i) in enumerate(self.create_list):
            self.AliveA[0, c] = 0
            for t in range(1, self.T):
                if not t - 1 in self.active_stages[i]:
                    self.AliveA[t, c] = self.AliveA[t - 1, c]

        for i in range(self.I):
            if not -1 in self.active_stages[i]:  # src node
                self.AliveT[0, i] = 0
            for t in range(1, self.T):
                if not t - 1 in self.active_stages[i]:
                    self.AliveT[t, i] = self.AliveT[t - 1, i]

        for t in range(self.T):
            for i in range(self.Cr):
                if self.create_list[i][0] not in self.krange(t):
                    self.create[t, i] = 0
            for i in range(self.De):
                if self.delete_list[i][0] not in self.krange(t):
                    self.delete[t, i] = 0
            if self.single_fwd:
                for d in range(self.De):
                    (k, i) = self.delete_list[d]
                    if i in self.protected_indices:
                        continue
                    if k == max(self.han_deps[i] + self.han_users[i]) and k == t:
                        self.delete[t, d] = 1
                        pass

    def solve(self, solver=""):
        # some solvers have no support of 'Time limit reached' status
        print(f"Nb comp: {len(self.Comp)}, T:{self.T}")
        print(f"time limit {self.ilp_solver_params['TimeLimit']}")
        try:
            solver = get_solver(
                solver, msg=0, timeLimit=self.ilp_solver_params["TimeLimit"]
            )
        except:
            avail_solver = listSolvers(onlyAvailable=True)[0]
            #     print(f"Cannot get {solver}, will use {avail_solver}")
            solver = get_solver(
                avail_solver, msg=0, timeLimit=self.ilp_solver_params["TimeLimit"]
            )
        # print("start solving")
        last_time = time.time()
        status = self.md.solve(solver)
        time_taken = time.time() - last_time
        clean_time_taken = time.strftime("%H:%M:%S", time.gmtime(time_taken))
        self.solving_time = clean_time_taken
        self.status = LpStatus[status]  # readable status
        self.feasible = status == 1

        sol = self.sol
        if self.feasible:
            # print(f"finished solving in {self.md.solutionTime}")
            # print(f"objective {self.md.objective.value()}")
            self.solve_time = self.md.solutionTime
            print(f"Solved in {self.solve_time}s")
            self.active_steps = []
            for t in list(range(self.loss_idx + 1, self.T)) + list(
                range(self.loss_idx + 1)
            ):
                for k in self.krange(t):
                    if not sol(self.sumComp[t, k]):
                        continue
                    self.active_steps.append((t, k))

    def sol(self, value):
        if hasattr(value, "value"):
            return value.value() > 0.9999
        return value > 0.9999



from psutil import virtual_memory
from typing import Any, Dict
from pulp import lpSum
import math
from rkgb.core.hierarchical import HierarchicalGraph
from .ilp_utils import RkLpVariable
from .ilp_model import ModelPULP


class ModelPULPOffload(ModelPULP):
    def __init__(
        self,
        hgraph: HierarchicalGraph,
        peak_budget: int,
        save_budget=None,
        ilp_solver_params: Dict[str, Any] = ...,
        gcd=None,
        accurate_mem=False,
        protected_names=...,
        grouping=True,
        grad_mode="free", #['free', 'accumulate', 'zero_grad']
        optimize_metrics=None,
        activation_offload=False,
        dynamic_batch_size=False,
        minor_offload_size=None,
        bandwidth=None,
        ram_budget=None,
        cpu_optimize=True,
        **kwargs
    ):
        super().__init__(
            hgraph,
            peak_budget,
            save_budget,
            ilp_solver_params,
            gcd,
            accurate_mem,
            protected_names,
        )

        self.with_grad = accurate_mem
        self.single_fwd = accurate_mem
        self.single_bwd = accurate_mem
        self.grouping = grouping
        self.grad_mode = grad_mode
        self.activation_offload = activation_offload
        self.optimize_metrics = optimize_metrics
        self.use_correction_term = True
        self.cpu_constant_cost = 100
        self.dynamic_batch_size = dynamic_batch_size
        self.ram_budget = ram_budget or virtual_memory().available * 0.9/self.gcd
        self.cpu_optimize = cpu_optimize
        self.minor_offload_size = minor_offload_size
        self.bandwidth = bandwidth
        if not self.optimize_metrics:
            raise UserWarning("Trying to solve ILP with offload but "\
                              "optimization method is not provided."\
                              "This is not implemented yet.")

        if self.grad_mode != "free":
            raise UserWarning("Parameter gradients are not freed during backward, "\
                              "which is not implemented in ILP.")

    def build(self):
        # OVERWRITTING METHOD
        self.config_offload()
        super().add_variables()
        self.add_offload_param_variables()
        if self.activation_offload:
            self.add_offload_activation_variables()
            self.add_offload_activation_constraints()
        super().add_constraints()
        self.add_offload_param_constraints()
        self.add_ram_constraints()
        # super().add_objective()
        self.add_offload_objective()
        self.add_offload_time_constraints()

    def config_offload(self):
        if self.optimize_metrics:
            optimize_metrics = self.optimize_metrics
            self.optimizer_states_factor = optimize_metrics[
                "optimizer_states_factor"
            ]  # *weight size
            self.cpu_optimize_speed = (
                optimize_metrics["cpu_optimize_speed"] / self.gcd
            )  # B/ms
            self.gpu_optimize_speed = (
                optimize_metrics["gpu_optimize_speed"] / self.gcd
            )  # B/ms
            self.optimizer_overhead_factor = optimize_metrics[
                "optimizer_overhead"
            ]  # *weight size
        else:
            self.optimizer_overhead_factor = 0
            self.optimizer_states_factor = 0
            self.cpu_optimize_speed = 0
            self.gpu_optimize_speed = 0

        # self.bandwidth = optimize_metrics["bandwidth"]  # bandwidth
        self.bandwidthOfl = self.bandwidth / self.gcd  # byte/ms
        self.bandwidthPrf = self.bandwidth / self.gcd  # byte/ms
        
        self.schedule_offload_time = []
        self.schedule_prefetch_time = []
        self.schedule_fwd_wait_time = []
        self.schedule_bwd_wait_time = []
        for list_sched in self.list_list_sched:
            self.schedule_offload_time.append([])
            self.schedule_prefetch_time.append([])
            self.schedule_fwd_wait_time.append([])
            self.schedule_bwd_wait_time.append([])
            for sched in list_sched:
                self.schedule_offload_time[-1].append(sched.offload_mem/self.bandwidthOfl / self.gcd)
                self.schedule_prefetch_time[-1].append(sched.prefetch_mem/self.bandwidthPrf / self.gcd)
                self.schedule_fwd_wait_time[-1].append(sched.fwd_wait_time)
                self.schedule_bwd_wait_time[-1].append(sched.bwd_wait_time)

        self.param2hcn = dict()
        self.parameters = []
        for k, v in self.hgraph.parameter_groups.items():
            self.param2hcn[len(self.parameters)] = k
            self.parameters.append(v)

        self.hcn2param = {t: [] for t in range(self.T)}

        for p, hcn_s in self.param2hcn.items():
            for hcn in hcn_s:
                self.hcn2param[hcn].append(p)

        self.parameter_size = [
            sum(pnode.mem for pnode in p) / self.gcd for p in self.parameters
        ]
        self.parameter_gradient_size = [
            sum(pnode.mem for pnode in p if pnode.info.requires_grad) / self.gcd
            for p in self.parameters
        ]
        self.W = len(self.parameters)

    def add_offload_time_constraints(self):
        for t in range(self.T):
            for k in self.krange(t):
                self.md += self.Time[t, k] >= self.time_step_prefetch(t, k)
                self.md += self.Time[t, k] >= (
                    self.time_step_offload(t, k) + self.time_step_optimize_self(t, k)
                )
                self.md += self.Time[t, k] >= self.time_step_optimize(t, k)
                self.md += self.Time[t, k] >= (
                    self.time_step_compute(t, k)
                    + self.time_step_optimize_self(t, k, cpu=False)
                    + self.time_step_offload_self(t, k)
                    + self.time_step_optimize_self(t, k, cpu=True)
                )

                self.md += self.Time[t, k] >= (
                    self.time_step_compute(t, k)+
                    self.time_step_prefetch_self(t,k))

    def add_offload_activation_variables(self):
        """
        We assume single forwarard/backward mode for ILP, thus only
        one possible source/user for phantom activations.
        """
        phantom_idx = [
            (t, k, j)
            for t in range(self.T)
            for k in self.krange(t)
            for j in range(self.J)
        ]
        # at step (t,k), removal of phantom of the j-th node.
        # self.RemovalP = RkLpVariable.dicts("RemovalP", phantom_idx)
        self.OflP = RkLpVariable.dicts("OflP", phantom_idx, lowBound=0, upBound=None)
        self.PrfP = RkLpVariable.dicts("PrfP", phantom_idx, lowBound=0, upBound=None)
        # pass
        self.phantoms = {}
        for j in range(self.J):
            for o, sched in enumerate(self.list_list_sched[j]):
                sub_cluster = self.sub_clusters[j]
                phantoms = [
                    sub_cluster.translate_representee_node(re_anode)
                    for re_anode in sched.phantoms
                    if (
                        re_anode.allocation_type == "data"  # Only support data offload
                        and re_anode.mem > self.minor_offload_size
                    )
                ]
                self.phantoms[(j, o)] = phantoms

    def add_offload_param_variables(
        self,
    ):
        lb = 0 if self.dynamic_batch_size else 1
        self.req_w = RkLpVariable("Required_w", lowBound=lb, upBound=1, cat="Continuous")


        param_idx = [
            (t, k, w)
            for t in range(self.T)
            for k in self.krange(t)
            for w in range(self.W)
        ]

        self.AliveW = RkLpVariable.dicts(
            "AliveW", param_idx
        )  # parameter w is alive at the start of step j.
        self.AliveG = RkLpVariable.dicts(
            "AliveG", param_idx
        )  # w.grad is alive at the start of step j.
        self.AliveO = RkLpVariable.dicts(
            "AliveO", param_idx
        )  # w.grad is alive at the start of step j.
        self.OflW = RkLpVariable.dicts("OflW", param_idx)  # Offload weight
        self.OflG = RkLpVariable.dicts("OflG", param_idx)  # Offload gradient
        self.PrfW = RkLpVariable.dicts("PrfW", param_idx)  # Prefetch gradient
        self.PrfG = RkLpVariable.dicts("PrfG", param_idx)  # Prefetch gradient
        self.OflO = RkLpVariable.dicts("OflO", param_idx, 
                        upBound=1 if self.optimize_metrics else 0)  # Offload optimizer states
        self.PrfO = RkLpVariable.dicts("PrfO", param_idx, 
                        upBound=1 if self.optimize_metrics else 0)  # Prefetch optimizer states

        if self.cpu_optimize and self.optimize_metrics:
            self.OptC = RkLpVariable.dicts("OptC", param_idx)  # Optimize on cpu
        else:
            self.OptC = RkLpVariable.dicts("OptC", param_idx, lowBound=0, upBound=0)
        self.sumOptC = dict()
        for w in self.param2hcn:
            self.sumOptC[w] = lpSum(
                self.OptC[t, k, w] for t in range(self.T) for k in self.krange(t)
            )

        self.param_grad_mem = {(t, k): 0 for t in range(self.T) for k in self.krange(t)}

        self.prefill_offload()

    def accumC_grad(self, w):
        # if grad_accumulation, gradient stored on CPU from previous iterations
        if self.grad_mode == "free":
            return self.sumOptC[w]


    def accumC_optimizer_states(self, w):
        # if grad_accumulation, optimizer states stored on CPU from previous iterations
        return self.accumC_grad(w)

    def max_OflGProg(self, t, k, w):
        return self.OflGProg[t, k, w] + (
            self.OflWProg[t, k, w] * (self.grad_mode == "free") * self.w_by_wg(w)
        )

    def w_by_wg(self, w):
        if self.parameter_gradient_size[w] == 0:
            return self.gcd
        return self.parameter_size[w] / self.parameter_gradient_size[w]

    def activation_mem(self, t, k):
        return self.U[t, k] - self.removal_phantom_mem(t, k)

    def save_mem(self, t, k):
        # OVERWRITTING METHOD
        return self.activation_mem(t, k) + self.all_param_mem(t, k)

    def overhead_mem(self, t, k):
        return self.act_overhead(t, k) + self.param_overhead(t, k)

    def param_overhead(self, t, k):
        optimizer_overhead = 0
        if k > self.loss_idx:  # and k in self.hcn2param:
            l_w = self.hcn2param[k]
            optimizer_overhead += sum(
                (self.req_w - self.sumOptC[w])
                * self.parameter_gradient_size[w]
                * self.optimizer_overhead_factor
                for w in l_w
            )
        return [optimizer_overhead]

    def all_param_mem(self, t, k, with_multiplier=True):
        return (
            self.parameter_mem(t, k)
            + self.param_grad_mem[t, k]
            + self.optimizer_states_mem(t, k)
            + (1 - self.req_w) * self.peak_budget * with_multiplier
        )

    def removal_phantom_mem(self, t, k):
        if not self.activation_offload:
            return 0
        mem = 0
        for j in range(self.J):
            mem += self.OflPProg[t, k, j] - self.PrfPProg[t, k, j]
        return mem

    def parameter_mem(self, t, k):
        parameter_mem = lpSum(
            (self.AliveW[t, k, w] + self.PrfW[t, k, w]) * self.parameter_size[w]
            for w in range(self.W)
        )
        return parameter_mem

    # def param_grad_mem(self,t, k):
    #     grad_mem = lpSum(
    #         self.AliveG[t, k, w]
    #         * self.parameter_gradient_size[w]
    #         for w in range(self.W)
    #     )
    #     return grad_mem

    def optimizer_states_mem(self, t, k, with_overhead=True):
        optimizer_states_mem = lpSum(
            (
                (self.AliveO[t, k, w] + self.PrfO[t, k, w])
                * self.parameter_gradient_size[w]
                * self.optimizer_states_factor
            )
            for w in range(self.W)
        )
        return optimizer_states_mem

    def prefill_offload(
        self,
    ):
        
        for t in range(self.T):
            for k in range(self.T):
                for w in range(self.W):
                    if k not in self.krange(t):
                        self.PrfW[t, k, w] = 0
                        self.OflW[t, k, w] = 0

                    # fwd_i, bwd_i = self.param2hcn[w]
                    fwd_i = min(self.param2hcn[w])
                    bwd_i = max(self.param2hcn[w])
                    if k not in self.krange(t) or (t > fwd_i and t < bwd_i):
                        self.OptC[t, k, w] = 0

        if not self.grad_mode == "accumulate":
            for k in self.PrfG:
                self.PrfG[k] = 0

        if not self.optimizer_states_factor:
            for t, k, w in self.AliveO:
                self.AliveO[(t, k, w)] = self.req_w - self.accumC_optimizer_states(w)
                self.PrfO[(t, k, w)] = 0
                self.OflO[(t, k, w)] = 0
        if self.grad_mode in ["free"]:
            # If gradient is freed ASAP, OflG will be merged into OflW
            for t, k, w in self.OflG:
                self.OflG[(t, k, w)] = 0
            for t, k, w in self.AliveG:
                grad_size = self.parameter_gradient_size[w]
                if grad_size == 0:
                    self.AliveG[(t, k, w)] = 0
                    continue
                if len(self.param2hcn[w]) <= 2:
                    self.AliveG[(t, k, w)] = 0
                    if k == max(self.param2hcn[w]):
                        # self.overhead[k] = [v+grad_size for v in self.overhead[k]]
                        self.param_grad_mem[t, k] += grad_size * self.req_w
                else:  # shared weight
                    bwd_first = min(x for x in self.param2hcn[w] if x > self.loss_idx)
                    bwd_last = max(self.param2hcn[w])
                    if t <= bwd_first or t > bwd_last:  # assume single bwd
                        self.AliveG[(t, k, w)] = 0
                    else:
                        self.AliveG[(t, k, w)] = 1
                        if k in self.param2hcn[w] and k > bwd_first:
                            # self.overhead[k] = [v+grad_size for v in self.overhead[k]]
                            self.param_grad_mem[t, k] += grad_size * self.req_w
        elif self.grad_mode in ["accumulate"]:
            for t, k, w in self.AliveG:
                bwd_indices = [x for x in self.param2hcn[w] if x > self.loss_idx]
                if t in bwd_indices:
                    self.AliveG[(t, k, w)] = 1
                # if k == max(self.param2hcn[w]):
                #     self.overhead[k] = [v+grad_size for v in self.overhead[k]]
                # TODO: add offload gradient variables for gradient accumulation

        for t in range(self.T):
            for k in self.krange(t):
                self.param_grad_mem[t, k] += lpSum(
                    self.AliveG[t, k, w] * self.parameter_gradient_size[w]
                    for w in range(self.W)
                )

    def add_offload_param_constraints(
        self,
    ):
        # if with_grad, AliveG is a variable
        self.OflWProg = dict()
        self.OflGProg = dict()
        self.OptCProg = dict()
        self.PrfWProg = dict()
        self.PrfGProg = dict()
        self.OflOProg = dict()
        self.PrfOProg = dict()

        def get_progress(op, t, k, w):
            bwd_i = max(self.param2hcn[w])
            if bwd_i < t:  # after bwd of w
                progress = (
                    lpSum(op[t, kk, w] for kk in self.krange(t) if kk < k)
                    + lpSum(
                        op[tt, kk, w]
                        for tt in range(bwd_i + 1, t)
                        for kk in self.krange(tt)
                    )
                    + lpSum(
                        op[bwd_i, kk, w] for kk in self.krange(bwd_i) if kk >= bwd_i
                    )  # offload right after bwd_i
                )
            else:
                progress = (
                    lpSum(op[t, kk, w] for kk in self.krange(t) if kk < k)
                    + lpSum(op[tt, kk, w] for tt in range(t) for kk in self.krange(tt))
                    + lpSum(
                        op[tt, kk, w]
                        for tt in range(bwd_i + 1, self.T)
                        for kk in self.krange(tt)
                    )
                    + lpSum(
                        op[bwd_i, kk, w] for kk in self.krange(bwd_i) if kk >= bwd_i
                    )
                )
            return progress

        for t in range(self.T):
            for k in self.krange(t):
                t_, k_ = self.next_idx(t, k)
                # self.md += self.Time[t, k] >= self.time_step_prefetch(t, k)
                # self.md += self.Time[t, k] >= (self.time_step_offload(t, k)
                #                            + self.time_step_optimize_self(t, k))
                # self.md += self.Time[t, k] >= self.time_step_optimize(t, k)
                # self.md += self.Time[t, k] >= (self.time_step_compute(t, k)
                #                            + self.time_step_optimize_self(t, k, cpu=False)
                #                            + self.time_step_offload_self(t, k)
                #                            + self.time_step_optimize_self(t, k, cpu=True))

                for w in range(self.W):
                    self.PrfWProg[t, k, w] = get_progress(self.PrfW, t, k, w)
                    self.PrfGProg[t, k, w] = get_progress(self.PrfG, t, k, w)
                    self.OflWProg[t, k, w] = get_progress(self.OflW, t, k, w)
                    self.OflGProg[t, k, w] = get_progress(self.OflG, t, k, w)
                    self.OflOProg[t, k, w] = get_progress(self.OflO, t, k, w)
                    self.PrfOProg[t, k, w] = get_progress(self.PrfO, t, k, w)
                    self.OptCProg[t, k, w] = get_progress(self.OptC, t, k, w)

                    self.md += self.AliveW[t, k, w] <= self.req_w
                    self.md += self.OflWProg[t, k, w] <= self.req_w
                    self.md += self.OflGProg[t, k, w] <= self.accumC_grad(w)
                    self.md += self.OptCProg[t, k, w] <= self.max_OflGProg(t, k, w)
                    self.md += self.OptCProg[t, k, w] <= self.PrfWProg[
                        t, k, w
                    ] * self.w_by_wg(w)

                    self.md += self.AliveW[t, k, w] + self.OflWProg[
                        t, k, w
                    ] >= self.fraction_constant_param(
                        w
                    ) + self.fraction_instant_updated_param(
                        w
                    ) * (1/self.w_by_wg(w))
                    self.md += self.AliveG[t, k, w] + self.OflGProg[
                        t, k, w
                    ] >= self.fraction_remaining_gradients(w)

                    to_be_prefetched = self.parameter_size[w] * (
                        self.req_w - self.AliveW[t_, k_, w] - self.PrfW[t_, k_, w]
                    )
                    to_be_optimized = self.parameter_gradient_size[w] * (
                        self.sumOptC[w] - self.OptCProg[t, k, w]
                    )
                    self.md += to_be_prefetched >= to_be_optimized

                    diffW = self.AliveW[t_, k_, w] - self.AliveW[t, k, w]
                    self.md += diffW <= self.PrfW[t, k, w]

                    diffG = self.AliveG[t_, k_, w] - self.AliveG[t, k, w]
                    self.md += (
                        diffG
                        <= 1
                        * (
                            k in self.param2hcn[w] and k > self.loss_idx
                        )  # backward operations
                        + self.PrfG[t, k, w]
                    )

                    diffO = self.AliveO[t_, k_, w] - self.AliveO[t, k, w]
                    self.md += diffO <= self.PrfO[t, k, w]

                    self.md += self.OflOProg[t, k, w] >= self.PrfOProg[t, k, w]
                    self.md += self.AliveO[t, k, w] + self.OflOProg[
                        t, k, w
                    ] >= self.fraction_gpu_optimized(w)
                    self.md += self.OflOProg[t, k, w] <= self.fraction_gpu_optimized(w)

        # parameters and their user operations
        for w in self.param2hcn:
            fwd_i = min(self.param2hcn[w])
            bwd_i = max(self.param2hcn[w])
            self.md += self.PrfWProg[fwd_i, fwd_i, w] >= self.sumOptC[w]
            self.md += self.req_w - self.sumOptC[w] <= self.AliveO[bwd_i, bwd_i, w]
            if self.grad_mode == "accumulate":
                self.md += self.OflGProg[bwd_i, bwd_i, w] == self.accumC_grad(w)
                self.md += (
                    self.PrfGProg[bwd_i, bwd_i, w]
                    == self.accumC_grad(w) - self.sumOptC[w]
                )
            t_, k_ = self.next_idx(bwd_i, bwd_i)
            # self.md += self.AliveO[t_, k_, w] - self.AliveO[bwd_i, bwd_i, w] <= 1 - self.grad_mode == "accumulate"
            # self.md += self.AliveG[t_, k_, w] - self.AliveG[bwd_i, bwd_i, w] <= 1# - self.grad_mode == "accumulate"
            for t in range(self.T):
                for k in self.param2hcn[w]:
                    if k not in self.krange(t):
                        continue
                    self.md += (
                        self.sumComp[t, k] - 1 <= self.AliveW[t, k, w] - self.req_w
                    )

    def add_offload_activation_constraints(self):
        self.PrfPProg = dict()
        self.OflPProg = dict()

        def get_progress_phantom(op, t, k, j):
            fwd_i = min(self.sub_c2hcn[j])
            bwd_i = max(self.sub_c2hcn[j])

            if bwd_i < t or fwd_i > t:  # after bwd of w
                return 0

            progress = (
                lpSum(op[t, kk, j] for kk in self.krange(t) if kk < k)
                + lpSum(
                    op[tt, kk, j]
                    for tt in range(fwd_i + 1, t)
                    for kk in self.krange(tt)
                )
                + lpSum(op[fwd_i, kk, j] for kk in self.krange(fwd_i) if kk >= fwd_i)
            )

            return progress

        for j in range(self.J):
            fwd_i = min(self.sub_c2hcn[j])
            bwd_i = max(self.sub_c2hcn[j])
            saved_mem = lpSum(
                # self.saved_mem[j][o]*
                sum(anode.mem for anode in self.phantoms[j, o])
                / self.gcd
                * self.Comp[fwd_i, fwd_i, o]
                for o, sched in enumerate(self.list_list_sched[j])
            )
            for t in range(self.T):
                for k in self.krange(t):
                    if t >= fwd_i and t <= bwd_i:
                        self.PrfPProg[t, k, j] = get_progress_phantom(
                            self.PrfP, t, k, j
                        )
                        self.OflPProg[t, k, j] = get_progress_phantom(
                            self.OflP, t, k, j
                        )

                        # removal cannot be higher than phantom created during fwd
                        self.md += self.OflPProg[t, k, j] <= saved_mem
                    else:
                        self.PrfPProg[t, k, j] = 0
                        self.OflPProg[t, k, j] = 0

            self.md += self.PrfPProg[bwd_i, bwd_i, j] == self.OflPProg[bwd_i, bwd_i, j]

    def add_ram_constraints(self):
        self.md += self.ram_usage() <= self.ram_budget

    def ram_usage(self):
        mem_usage = 0
        # activation on cpu
        for t in range(self.loss_idx):
            for k in self.krange(t):
                j = self.hcn2sub_c[k]
                mem_usage += lpSum(
                    self.Comp[t, k, o] * self.list_list_sched[k][o].offload_mem/self.gcd
                    for o in range(self.nSched[k])
                )
        for w in range(self.W):
            bwd_i = max(self.param2hcn[w])
            mem_usage += self.parameter_size[w] * self.OflWProg[bwd_i, bwd_i, w]
            mem_usage += (self.parameter_gradient_size[w] 
                          * self.OflOProg[bwd_i, bwd_i, w]
                          * self.optimizer_states_factor)
            mem_usage += (self.parameter_gradient_size[w] 
                          * self.sumOptC[w] 
                          * (self.optimizer_states_factor+1))
            
        #parameters are already in RAM
        mem_usage -= sum(self.parameter_size)
        return mem_usage
        

    def add_offload_objective(self, bandwidth_cost=0.01):
        prf_cost = bandwidth_cost * lpSum(
            (self.PrfW[t, k, w] + self.PrfG[t, k, w] + self.PrfO[t, k, w])
            * self.parameter_size[w]
            / self.bandwidthPrf
            for t in range(self.T)
            for k in self.krange(t)
            for w in range(self.W)
        )
        ofl_cost = bandwidth_cost * lpSum(
            (self.OflW[t, k, w] + self.OflG[t, k, w] + self.OflO[t, k, w])
            * self.parameter_size[w]
            / self.bandwidthOfl
            for t in range(self.T)
            for k in self.krange(t)
            for w in range(self.W)
        )
        self.md += (
            lpSum(self.Time[t, k] for t in range(self.T) for k in self.krange(t))
            + prf_cost
            + ofl_cost
        )

    """
    The following functions are meant to be used for building constraints with
    meaningful names. They are not supposed to be called outside this file.
    By default, index t is for stage, k for step and w for the parameter group.
    """

    def fraction_gpu_optimized(self, w):
        return self.req_w - self.sumOptC[w]

    def fraction_constant_param(self, w):
        return 1 - self.parameter_gradient_size[w] / self.parameter_size[w]

    def fraction_instant_updated_param(self, w):
        if self.grad_mode == "accumulate":
            # TODO: not fully supported now
            return 0
        if self.grad_mode == "free":
            return self.req_w
        return self.req_w - self.sumOptC[w]

    def fraction_remaining_gradients(self, w):
        if self.grad_mode == "accumulate":
            # TODO: not fully supported now
            return self.req_w
        if self.grad_mode == "free":
            return 0

    # def fraction_after_bwd(self,w):
    #     return self.fraction_constant_param(w) + self.fraction_instant_updated_param(w)

    def time_step_offload(self, t, k):
        j = self.hcn2sub_c[k]
        sub_sched_time = lpSum(
            self.Comp[t, k, o] * self.schedule_offload_time[j][o]
            for o in range(self.nSched[k])
        )
        mem = 0
        for w in range(self.W):
            mem += self.parameter_size[w] * self.OflW[t, k, w]
            optim_state = self.optimizer_states_factor * self.OflO[t, k, w]
            mem += self.parameter_gradient_size[w] * (self.OflG[t, k, w] + optim_state)
        if self.activation_offload:
            for j in range(self.J):
                mem += self.OflP[t, k, j]

        return mem / self.bandwidthOfl + sub_sched_time

    def time_step_offload_self(self, t, k):
        j = self.hcn2sub_c[k]
        sub_sched_time = lpSum(
            self.Comp[t, k, o] * self.schedule_fwd_wait_time[j][o]
            for o in range(self.nSched[k])
        )
        mem = 0
        for w in self.hcn2param[k]:
            mem += self.parameter_size[w] * self.OflW[t, k, w]
            optim_state = self.optimizer_states_factor * self.OflO[t, k, w]
            mem += self.parameter_gradient_size[w] * (self.OflG[t, k, w] + optim_state)
        if self.activation_offload:
            j = self.hcn2sub_c[k]
            if j is not None:
                mem += self.OflP[t, k, j]

        return mem / self.bandwidthOfl + sub_sched_time

    def time_step_prefetch(self, t, k):
        j = self.hcn2sub_c[k]
        sub_sched_time = lpSum(
            self.Comp[t, k, o] * self.schedule_prefetch_time[j][o]
            for o in range(self.nSched[k])
        )

        mem = 0
        for w in range(self.W):
            mem += self.parameter_size[w] * self.PrfW[t, k, w]
            optim_state = self.optimizer_states_factor * self.PrfO[t, k, w]
            mem += self.parameter_gradient_size[w] * (self.PrfG[t, k, w] + optim_state)
        if self.activation_offload:
            for j in range(self.J):
                mem += self.PrfP[t, k, j]

        return mem / self.bandwidthPrf + sub_sched_time

    def time_step_prefetch_self(self, t, k):
        j = self.hcn2sub_c[k]
        sub_sched_time = lpSum(
            self.Comp[t, k, o] * self.schedule_bwd_wait_time[j][o]
            for o in range(self.nSched[k])
        )
        mem = 0
        for w in self.hcn2param[k]:
            mem += self.parameter_size[w] * self.PrfW[t, k, w]
            optim_state = self.optimizer_states_factor * self.PrfO[t, k, w]
            mem += self.parameter_gradient_size[w] * (self.PrfG[t, k, w] + optim_state)

        return mem / self.bandwidthPrf + sub_sched_time

    def time_step_optimize(self, t, k, cpu=True):
        """
        We assume that GPU optimization happens only right after
        the backward, which should be accessed from time_step_optimize_self.
        """
        if not self.optimize_metrics:
            return 0
        mem = 0
        for w in range(self.W):
            mem += self.parameter_gradient_size[w] * self.OptC[t, k, w]
        return mem / self.cpu_optimize_speed + self.cpu_constant_cost * self.req_w

    def time_step_optimize_self(self, t, k, cpu=True):
        """
        We assume that GPU optimization happens only right after
        the backward, but CPU optimization could happen anytime and
        represented by self.OptC.
        """
        if not self.optimize_metrics:
            return 0
        mem = 0
        if cpu:
            for w in self.hcn2param[k]:
                opt_fraction = self.OptC[t, k, w]
                mem += self.parameter_gradient_size[w] * opt_fraction
            return mem / self.cpu_optimize_speed

        elif k <= self.loss_idx:  # forward
            return 0
        else:
            for w in self.hcn2param[k]:
                opt_fraction = self.fraction_instant_updated_param(w)
                mem += self.parameter_gradient_size[w] * opt_fraction
            return mem / self.gpu_optimize_speed

    def time_step_compute(self, t, k):
        return lpSum(self.Comp[t, k, o] * self.time[k][o] for o in range(self.nComp[k]))



# This file is not actually used after updating rk_rotor.py
# ==========================
# definition file of RK_Chain
# also contains RK_Chain builder -> depends on use_chk.py
#  based on rotor/algorithms/parameters.py
# ==========================
from rkgb.core.hierarchical import HierarchicalGraph, HierarchicalCluster
from ..main import get_sched, translate
from ...op_schedule import OpSchedule, DeleteOp, ComputeOp
from . import def_sequence as rkseq

import math
from copy import deepcopy

# ==========================
# ======== RK Block ========
# ==========================


def forward_schedule(op_list, cluster):
    return OpSchedule(op_list, loss_idx=len(op_list) - 1, cluster=cluster, correct_overhead=False)
def backward_schedule(op_list, cluster):
    return OpSchedule(op_list, loss_idx=0, cluster=cluster, correct_overhead=False)

def get_sched_peak_with_interfaces(sched, only_start=False, only_end=False):
    assert not (only_start and only_end)
    start_idx = sched.loss_idx+1 if only_end else None
    end_idx = sched.loss_idx if only_start else None
    return max(sched.save_mem_with_interfaces[start_idx:end_idx] + sched.overhead[start_idx:end_idx])

def get_oplist_peak(op_list, cluster):
    sched = OpSchedule(op_list, loss_idx=len(op_list) - 1, cluster=cluster, correct_overhead=False)
    return max(sched.save_mem_with_interfaces + sched.overhead)

def get_oplist_time(op_list):
    return sum(op.time for op in op_list if not op.disabled)

class RK_Block_Solution:
    # A solution contains a fwd_sched and a bwd_sched
    # and also size_a_bar, time_fwd, time_bwd, overhead_fwd, overhead_bwd

    # prologue_op_list is added to the forward part of op_sched. Can be []
    # pred_disable is a predicate for ops that should be disabled from the schedules.
    #              Used to disable delete operations about the input data
    def __init__(self, block, sub_cluster, op_sched, prologue_op_list, pred_disable):
        
        def translate_and_filter(op_list):
            translated = translate(sub_cluster, op_list)
            return [ op for op in translated if not pred_disable(op) ]

        self.fwd_op_list = translate_and_filter(prologue_op_list + op_sched.op_list[: op_sched.loss_idx])
        self.bwd_op_list = translate_and_filter(op_sched.op_list[op_sched.loss_idx + 1 :])


        full_sched = OpSchedule(
            self.fwd_op_list
            + [ComputeOp(sub_cluster.loss_cnode)]
            + self.bwd_op_list,
            loss_idx=len(self.fwd_op_list),
            cluster=sub_cluster,
            correct_overhead=False,
        )

        # Include in abar size the size of output Tensors which are alive when computing loss
        self.size_a_bar = full_sched.mem + block.mem_output
        assert (block.mem_output == sum(anode.mem for anode in full_sched.interfaces["output_data_anodes"]
                                        if full_sched.alive_list[full_sched.loss_idx][anode.name]))

        self.time_fwd = full_sched.fwd_time
        self.time_bwd = full_sched.bwd_time
        # for rotor, overhead needs to be equal to peak - size of input and output.
        #     for backward, input is [input + grad of output + output with phantoms] and output is [grad of input]
        self.overhead_fwd = get_sched_peak_with_interfaces(full_sched, only_start=True) - block.mem_input - self.size_a_bar
        self.overhead_bwd = get_sched_peak_with_interfaces(full_sched, only_end=True) - 2 * block.mem_input - self.size_a_bar - block.mem_output

class RK_Block:
    ## Build a block from a Compute node, or several compute nodes merged
    ## if the first ones have no BWD.
    def __init__(self, h_compute_nodes):
        assert len(h_compute_nodes) >= 1
        previous_nodes, final_node = h_compute_nodes[:-1], h_compute_nodes[-1]
        assert all(hcn.sub_cluster is None for hcn in previous_nodes)

        # Since we have a sequential structure, we assume only the first compute node of the prologue has inputs
        input_anode_data = list(han.anode for han in h_compute_nodes[0].deps)
        output_anode_data = list(han.anode for han in h_compute_nodes[-1].users)
        self.mem_input = sum(anode.mem for anode in input_anode_data)
        self.mem_output = sum(anode.mem for anode in output_anode_data)

        prologue_op_list = []
        for f_hcn in previous_nodes:
            prologue_op_list += f_hcn.ff_op_list

        self.Fn_op_list = prologue_op_list + final_node.ff_op_list

        # Rotor assumes that input data is deleted only with Fn, so we remove delete operations from Fc 
        # and Fe (for Fe, we also prevent deletion of gradients of the input)
        def is_delete_of_input_data(op):
            return (isinstance(op, DeleteOp) and op.target in input_anode_data)

        def is_delete_of_input_data_or_gradient(op):
            return (isinstance(op, DeleteOp) and 
                    any(op.target.anode.main_target == anode.main_target for anode in input_anode_data))

        self.Fc_op_list = list(op for op in self.Fn_op_list if not is_delete_of_input_data(op))

        # One block solution for each FWD/BWD schedule
        self.sols = list(
            RK_Block_Solution(self, final_node.sub_cluster, op_sched, prologue_op_list, pred_disable=is_delete_of_input_data_or_gradient)
            for op_sched in get_sched(final_node.sub_cluster)
        )

        # TODO: measure both Fc and Fn peaks, they might be different
        self.overhead_fast_fwd = get_oplist_peak(self.Fc_op_list, final_node.sub_cluster) - self.mem_input - self.mem_output
        self.time_fast_fwd = get_oplist_time(self.Fc_op_list)

# ==========================
# ======== RK CHAIN ========
# ==========================

def make_list_of_empty_lists(n):
    return [[] for _ in range(n)]

class RK_Chain:
    def __init__(self, hg: HierarchicalGraph, mem_unit=1):
        self.body = []
        self.mem_unit = mem_unit
        loss_idx = hg.list_HCNs.index(hg.loss_hcn)

        no_grad_hcns = []
        for i, hcn in enumerate(hg.list_HCNs[:loss_idx]):  # toposorted
            # we only make a Block if the hcn has a BWD, otherwise we keep it to merge with the next one
            no_grad_hcns.append(hcn)
            if hcn.sub_cluster is not None:
                self.body.append(RK_Block(no_grad_hcns))
                no_grad_hcns = []
        

        mkl = make_list_of_empty_lists
        self.ln = len(self.body)
        nb = self.ln + 1
        self.fw = mkl(nb)
        self.bw = mkl(nb)
        self.cw = [None] * (nb + 1)
        self.cbw = mkl(nb + 1)
        self.fwd_tmp = mkl(nb)
        self.bwd_tmp = mkl(nb)
        self.ff_fwd_tmp = [None] * (nb)
        self.ff_fw = [None] * (nb)

        self.nb_sol = []
        for i, b in enumerate(self.body):
            self.nb_sol.append(len(b.sols))
            if self.nb_sol[-1] == 0:
                raise Exception(
                    f"We need at least one solution per block. "
                    f"Here {b.block_name} has no solution"
                )
            for sol in b.sols:
                self.fw[i].append(sol.time_fwd)
                self.bw[i].append(sol.time_bwd)
                self.cbw[i + 1].append(sol.size_a_bar)
                self.fwd_tmp[i].append(sol.overhead_fwd)
                self.bwd_tmp[i].append(sol.overhead_bwd)
            self.cw[i] = b.mem_input
            self.ff_fwd_tmp[i] = b.overhead_fast_fwd
            self.ff_fw[i] = b.time_fast_fwd

        self.cw[self.ln] = sum(value.mem for value in hg.output_data_HANs)  # the final output(s)

        # The final output is not counted in the budget, so we should remove it from 
        #  the abar sizes.
        self.cbw[-2] = [(c - self.cw[self.ln]) for c in self.cbw[-2]]

        # for the Loss block :
        self.nb_sol.append(1)
        self.fw[-1] = [0]
        self.bw[-1] = [0]
        self.cw[-1] = 0
        self.cbw[-1] = [0]
        self.fwd_tmp[-1] = [0]
        self.bwd_tmp[-1] = [0]
        self.ff_fwd_tmp[-1] = 0
        self.ff_fw[-1] = 0

        self.discretize()


    def discretize(self):
        def discretize_(values):
            return [int(math.ceil(value / self.mem_unit)) for value in values]

        self.cw = discretize_(self.cw)
        self.cbw = [discretize_(x) for x in self.cbw]
        self.fwd_tmp = [discretize_(x) for x in self.fwd_tmp]
        self.bwd_tmp = [discretize_(x) for x in self.bwd_tmp]
        self.ff_fwd_tmp = discretize_(self.ff_fwd_tmp)

    def _get_norecomputation_sequence(self):
        def cheapest_option(layer):
            return min(range(self.nb_sol[layer]), key=lambda i: self.fw[layer][i] + self.bw[layer][i])

        forward = []
        backward = []
        for i in range(self.ln):
            option = cheapest_option(i)
            forward.append(rkseq.SeqBlockFe(i, option, self.body[i].sols[option].fwd_op_list))
            backward.append(rkseq.SeqBlockBwd(i, self.body[i].sols[option].bwd_op_list))
        backward.reverse()
        return rkseq.RK_Sequence(forward + [rkseq.SeqLoss()] + backward )

    def _get_allrecomputation_sequence(self):
        def smallest_option(layer):
            return min(range(self.nb_sol[layer]), 
                       key=lambda i: self.cbw[layer+1][i]+max(self.fwd_tmp[layer][i], self.bwd_tmp[layer][i]))

        sequence = rkseq.RK_Sequence()

        for j in reversed(range(self.ln + 1)):
            sequence.insert(rkseq.SeqBlockFc(0, self.body[0].Fc_op_list))
            for i in range(1, j):
                sequence.insert(rkseq.SeqBlockFn(i, self.body[i].Fn_op_list))
            if j == self.ln:
                sequence.insert(rkseq.SeqLoss())
            else:
                option = smallest_option(j)
                sequence.insert(rkseq.SeqBlockFe(j, option, self.body[j].sols[option].fwd_op_list))
                sequence.insert(rkseq.SeqBlockBwd(j, self.body[j].sols[option].bwd_op_list))
        return sequence

    def get_min_memory(self):
        seq = self._get_allrecomputation_sequence()
        return self.simulate(seq, display=False) * self.mem_unit

    def get_max_memory(self):
        seq = self._get_norecomputation_sequence()
        return self.simulate(seq, display=False) * self.mem_unit

    def simulate(self, sequence, display=True, stopAtLoss=False):
        return sequence.simulate(self, display, stopAtLoss)

    def __str__(self):
        result = f"Chain of length {self.ln}\n"
        for i in range(self.ln):
            result += f"Block {i}: {self.cw[i]} {self.ff_fwd_tmp[i]} {self.ff_fw[i]:.2f} {self.nb_sol[i]}\n"
            for j in range(self.nb_sol[i]):
                peak_fw = self.cw[i] + self.cw[i+1] + self.fwd_tmp[i][j]
                peak_bw = self.cw[i] + self.cbw[i+1][j] + self.cw[i] + self.cw[i+1] + self.bwd_tmp[i][j]
                result += f"   Option {j}: {self.cbw[i+1][j]} {self.fwd_tmp[i][j]}:{peak_fw} {self.bwd_tmp[i][j]}:{peak_bw} {self.fw[i][j]:.2f} {self.bw[i][j]:.2f}\n"
        return result
